/*
 * Auth: Dan Quinlan
 *       Los Alamos National Labs
 * Desc: Functions added to free memory after using PARTI 
 *       as a means of tracing memory leaks in software using 
 *       PARTI.
 */

#include "addition.h"
#include "datamove.h"
#include "hash.h"
#include "utils.h"
#include <assert.h>
#include <stdlib.h>

void cleanup_after_PARTI() {
#if PARTI_DEBUG > 2
  printf("Inside of cleanup_after_PARTI!\n");
#endif
  free_message_buffers();
  cleanup_da_table();
  cleanup_subarray_table();
  cleanup_exch_table();
  pcheck();
}

void delete_XDARRAY(struct dArray_rec* X) {
#if PARTI_DEBUG > 2
  printf("In delete_XDARRAY: X->referenceCount = %d\n", X->referenceCount);
  printf("In delete_XDARRAY: X->decomp->referenceCount = %d\n", X->decomp->referenceCount);
#endif
  assert(X->referenceCount >= 0);

  if (X->referenceCount-- == 0) {
    if (X->left_ghostCells != NULL) free(X->left_ghostCells);
    X->left_ghostCells = NULL;
    X->right_ghostCells = NULL;
    X->dimVecG = NULL;
    X->dimVecL = NULL;
    X->dimVecL_L = NULL;
    X->dimVecL_R = NULL;
    X->g_index_low = NULL;
    X->g_index_hi = NULL;
    X->local_size = NULL;
    X->decompDim = NULL;
    X->decompPosn = NULL;
    if (X->dimDist != NULL) free(X->dimDist);
    X->dimDist = NULL;
    delete_XDECOMP(X->decomp);
    X->decomp = NULL;
    pfree("struct dArray_rec", sizeof(struct dArray_rec), X);
  }
}

void delete_DARRAY(DARRAY* X) {
#if PARTI_DEBUG > 2
  printf("In delete_DARRAY: X->referenceCount = %d\n", X->referenceCount);
  printf("In delete_DARRAY: X->decomp->referenceCount = %d\n", X->decomp->referenceCount);
#endif
  parti_free_darray(X);
}

void delete_SCHED(SCHED* X) {
  int i = 0;
#if PARTI_DEBUG > 2
  printf("In delete_SCHED: X->referenceCount = %d\n", X->referenceCount);
  printf("In delete_SCHED: X->ID_Number = %d\n", X->ID_Number);
#endif
  assert(X->referenceCount >= 0);

  if (X->referenceCount-- == 0) {
    for (i = 0; i < PARTI_numprocs(); i++) {
      if (X->rData[i] != NULL) free(X->rData[i]);
      X->rData[i] = NULL;
      if (X->sData[i] != NULL) free(X->sData[i]);
      X->sData[i] = NULL;
    }
    pfree("SCHED", sizeof(SCHED), X);
  }
}

void delete_XDECOMP(struct decomp_rec* X) {
#if PARTI_DEBUG > 2
  printf("In delete_XDECOMP: X->referenceCount = %d\n", X->referenceCount);
#endif
  assert(X->referenceCount >= 0);
  if (X->referenceCount-- == 0) {
    free(X->dimVec);
    X->dimVec = NULL;
    free(X->dimProc);
    X->dimProc = NULL;
    free(X->dimDist);
    X->dimDist = NULL;    
    pfree("struct decomp_rec", sizeof(struct decomp_rec), X);
  }
}

void delete_DECOMP(DECOMP* X) {
#if PARTI_DEBUG > 2
  printf("In delete_DECOMP: X->referenceCount = %d\n", X->referenceCount);
#endif
  parti_free_decomp(X);
}

void delete_VPROC(VPROC* X) {
  free(X->dimVec);
  X->dimVec = NULL;
  pfree("VPROC", sizeof(VPROC), X);
}
