/*
 * Auth: Alan Sussman (als@cs.umd.edu)
 *       University of Maryland, College Park
 * Desc: The original MultiBlockParti interface
 */

#ifndef BSPARTI_H
#define BSPARTI_H

#include "multiblockparti.h"

#define GRPSIZE   64
#define NTYPE1	  100

/* datatypes */
#define DECOMP parti_decomp_t
#define DARRAY parti_darray_t
#define SCHED parti_sched_t

/* describes a range of ghost cells to fill, or a block to move */
/* Rewritten : Gagan 07/05/93 */ 
struct range_rec {
  int rangeDims[MAX_DIM]; 
  int max[MAX_DIM]; 
  int min[MAX_DIM]; 
  int step[MAX_DIM];
  int indexVals[MAX_DIM];  /* the indices for each array dimension */
};

/* Rewritten: Gagan 07/05/93 */
struct dest {
  char name[10];
  int* ptr;
  int startPosn; 
  int step[MAX_DIM];
};

struct dArray_rec {
  int nDims;
  int* left_ghostCells;       /* number of left internal ghost cells in each dim */
  int* right_ghostCells;      /* number of right internal ghost cells in each dim */
  int* dimVecG;               /* total size of each dim */
  int* dimVecL;               /* local size of each dim for central pieces*/
  int* dimVecL_L;             /* local size of each dim for left most piece */
  int* dimVecL_R;             /* local size of each dim for right most piece */
  int* g_index_low;           /* lower global index on my processor */
  int* g_index_hi;            /* upper global index on my processor */
  int* local_size;            /* Local size on my processor */  
  int* decompDim;             /* dim of decomp to which each dim aligned 
			       * defines how array aligned to decomp
			       * used with decomp to inialize decompPosn
			       * and dimDist 
			       */
  int* decompPosn;            /* coordinate position of processor in 
			       * the decomposition to which it's bound
			       * in the multi-dimensional decomposition
			       * space
			       */
  char* dimDist;              /* type of distribution in each dim */
  struct decomp_rec* decomp;  /* decomposition to which processor bound */
  int referenceCount;
};

struct decomp_rec {
  int nDims;
  int nProcs;
  int baseProc;
  int* dimVec;         /* size of decomposition in each dim */
  int* dimProc;        /* num processors allocated to each dim */
  char* dimDist;       /* type of distribution in each dim */
  int referenceCount;
};

typedef struct vProc_rec {
  int nDims;
  int nProcs;
  int usedProcs;
  int* dimVec;
} VPROC;

struct range {
  int rangeDim;
  int max;
  int min;
  int stride;
};

struct exch_rec {
  /* for a fixed index */
  int nFixedIndices;
  int indexDim[MAX_DIM];
  int indexVals[MAX_DIM];
  /* for a range of indices */
  int numRanges;
  int rangeDim[MAX_DIM];
  struct range ranges[MAX_DIM];
};

struct time_rec {
  double total;
  double comp;
  double comm;
};

/* declaration primitives */
VPROC* vProc(int, int*);
DECOMP* create_decomp(int, int*);
void embed(DECOMP*, VPROC*, int, int);
void distribute(DECOMP*, char*);
int section(DECOMP*, int*, int*, int**);
int partition(DECOMP*, int, int*, int, int*);
DARRAY* align(DECOMP*, int, int*, int*, int*, int*, int*, int*, int*, int*);

/* loop bound adjustment primitives */
int lalbnd(DARRAY*, int, int, int);
int laubnd(DARRAY*, int, int, int);

/* communiction primitives */
SCHED* ghostFillSched(DARRAY*, int, int*);
SCHED* ghostFillSpanSched(DARRAY*, int, int*);
SCHED* ghostFillAllSched(DARRAY*);
SCHED* exchSched(DARRAY*, int, int);
SCHED* subArraySched(DARRAY*, DARRAY*, int, int*, int*, int*, int*, int*, int*, int*, int*);

#define cDataMove parti_data_move_char
#define iDataMove parti_data_move_int
#define fDataMove parti_data_move_float
#define dDataMove parti_data_move_double

/* memory management primitives */
void laSizes(DARRAY*, int*);
#define free_sched parti_free_sched
void remove_exch_scheds();
void remove_subarray_scheds();

/* miscellaneous primitives */
int gLBnd(DARRAY*, int);
int gUBnd(DARRAY*, int);
int globalToLocal(DARRAY*, int, int);
int globalToLocalWithGhost(DARRAY*, int, int);
int localToGlobal(DARRAY*, int, int);
int localToGlobalWithGhost(DARRAY*, int, int);

/* globals */
extern int FFlag;

#endif /* BSPARTI_H */
