

#include "datamove.h"
#include "hash.h"
#include "utils.h"
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

char *RmsgBuf[GRPSIZE], *SmsgBuf[GRPSIZE], *local_msgBuf;
int RmsgBuf_size[GRPSIZE], SmsgBuf_size[GRPSIZE], local_msgBuf_size;
static int msgbuf_done_init = 0;








void init_message_buffers() {
  int i;

  for (i = 0; i < GRPSIZE; ++i) {
    RmsgBuf[i] = 0;
    RmsgBuf_size[i] = 0;
    SmsgBuf[i] = 0;
    SmsgBuf_size[i] = 0;    
  }

  local_msgBuf = 0;
  local_msgBuf_size = 0;

  msgbuf_done_init = 1;
}


/****************************************************************
 *  Frees all message buffers, resetting sizes to 0
 *****************************************************************/
void free_message_buffers() {
  int i;
  
  for (i = 0; i < GRPSIZE; i++) {
    if (RmsgBuf[i] != 0)
      free(RmsgBuf[i]);
    RmsgBuf[i] = 0;
    RmsgBuf_size[i] = 0;

    if (SmsgBuf[i] != 0)
      free(SmsgBuf[i]);
    SmsgBuf[i] = 0;
    SmsgBuf_size[i] = 0;
  }
  
  if (local_msgBuf != 0)
    free(local_msgBuf);
  local_msgBuf = NULL;
  local_msgBuf_size = 0;
}


void parti_data_move_char(char* array, SCHED* sched, char* dest) {
  int proc, nproc, grp_size, Wprocs;
  int srcPrc, GSsrc, destPrc, GSdest;
  int sending, receiving;
  int i, lPck, lSrc, sizeR[GRPSIZE], sizeS[GRPSIZE];
  char *RbufPtr[GRPSIZE], *SbufPtr[GRPSIZE];
  SchedData *sendData, *recvData;
  int j, k[MAX_DIM], currdim, done;
  char *arrayPtr, *destPtr, *tempPtr[MAX_DIM];
  GID_TYPE sGid[GRPSIZE], rGid[GRPSIZE];
  int j1, allmesg, recv_done[GRPSIZE], lWait;  
  
  if (!msgbuf_done_init)
    init_message_buffers();
 
  if (array == 0 || sched == 0 || dest == 0)
    return;
  
  proc = PARTI_myproc();
  nproc = PARTI_numprocs();
  
  if (GRPSIZE > nproc) {
    grp_size = nproc;
  } else {
    grp_size = GRPSIZE;
  }
  
  Wprocs = (nproc/grp_size)*grp_size;
  
  /* Break 0...(nproc-1) loop into groups of size grp_size */
  for (srcPrc = 0; srcPrc < nproc; srcPrc += grp_size) {
    if (srcPrc >= Wprocs) {
      GSsrc = nproc - Wprocs;
    } else {
      GSsrc = grp_size;
    }    

    GSdest = grp_size;
    
    if ((proc < (srcPrc + GSsrc)) && (proc >= srcPrc)) {
      sending = 1;
    } else {
      sending = 0;
    }
    
    for (destPrc = 0; destPrc < nproc; destPrc += grp_size) {
      if (destPrc >= Wprocs) {
	GSdest = nproc - Wprocs;
      }

      if ((proc < (destPrc + GSdest)) && (proc >= destPrc)) {
	receiving = 1;
      } else {
	receiving = 0;
      }

      /* check receive allocation */
      if (receiving) {
        for (lPck = 0; lPck < GSsrc; ++lPck) {
          i = srcPrc + lPck;
          if (sched->rMsgSz[i] > 0) { 
            sizeR[lPck] = sched->rMsgSz[i]*sizeof(char);	    
            if (i != proc) {
              if (sizeR[lPck] > RmsgBuf_size[lPck]) {
		/* need a bigger message buffer */
		RmsgBuf[lPck] = (char*)realloc(RmsgBuf[lPck], sizeR[lPck]);
                if (RmsgBuf[lPck] == 0) {
                  fprintf(stderr, "Cannot allocate space for message buffers");
                  return;
                }
                RmsgBuf_size[lPck] = sizeR[lPck];
              }
	      RbufPtr[lPck] = (char*)RmsgBuf[lPck];	      
            } else {
              /* copy from the local message buffer */
              if (sizeR[lPck] > local_msgBuf_size) {
                /* need a bigger local message buffer */
		local_msgBuf = (char*)realloc(local_msgBuf, sizeR[lPck]);
                if (local_msgBuf == 0) {
                  fprintf(stderr, "Cannot allocate space for local message buffer");
		  return;
                }
                local_msgBuf_size = sizeR[lPck];
              }
              RbufPtr[lPck] = (char*)local_msgBuf;
	    } 
	  }
	}
      } 
      
      /* check send allocation */
      if (sending) {
        for (lPck = 0; lPck < GSdest; ++lPck) {
	  i = lPck + destPrc;
          if (sched->sMsgSz[i] > 0) {
            sizeS[lPck] = sched->sMsgSz[i]*sizeof(char);
            if (i != proc) {
	      /* prepare a send */
              if (sizeS[lPck] > SmsgBuf_size[lPck]) {
                /* need a bigger message buffer */
		SmsgBuf[lPck] = (char*)realloc(SmsgBuf[lPck], sizeS[lPck]);
                if (SmsgBuf[lPck] == 0) {
                  fprintf(stderr, "Cannot allocate space for message buffers");
                  return;
		}
                SmsgBuf_size[lPck] = sizeS[lPck];
              }
              SbufPtr[lPck] = (char*)SmsgBuf[lPck];
            } else { /* als - 5/19/92 */
	      /* prepare the on-processor copy */
              if (sizeS[lPck] > local_msgBuf_size) {
		/* need a bigger local message buffer */
		local_msgBuf = (char*)realloc(local_msgBuf, sizeS[lPck]);
                if (local_msgBuf == 0) {
                  fprintf(stderr, "Cannot allocate space for local message buffer");
	          return;
                }
                local_msgBuf_size = sizeS[lPck];
              }
              SbufPtr[lPck] = (char*)local_msgBuf;
            }
	    
	    /* pack sends */
            sendData = sched->sData[i];
	    arrayPtr = array + sendData->startPosn;	    
            assert(sendData->numDims <= MAX_DIM);
	    
	    /* set lower bounds for all loops - 0 here */
	    for (currdim = 0; currdim < sendData->numDims; ++currdim) {
	      k[currdim] = 0;
	      tempPtr[currdim] = arrayPtr;
	    }
	    
	    currdim = done = 0;
	    while (1) {
	      /* reset indices and bump and reset buffer pointers for loops that are done */
	      while (k[currdim] >= sendData->numelem[currdim]) {
		if (currdim == (sendData->numDims - 1)) {
		  done = 1;
		  break;
		}		
		k[currdim] = 0;
		++currdim;
		++k[currdim];
		tempPtr[currdim] += sendData->str[currdim];
		/* reset all inner loop buffer pointers */
		for (j = currdim-1; j >= 0; --j) {
		  tempPtr[j] = tempPtr[j+1];
		}
	      }

	      if (done)
		break;

	      /* inner loop code */
	      currdim = 0;
	      *SbufPtr[lPck] = *tempPtr[currdim];
	      ++SbufPtr[lPck];
	      tempPtr[currdim] += sendData->str[currdim];
	      ++k[currdim];
	    }
            
	    if (i != proc) {
	      /* Modified by Serge Koren for merge with P++ */
	      PARTI_isend(NTYPE1+proc, SmsgBuf[lPck], sizeS[lPck], i, &sGid[lPck]);
	    }
	  }
	}
      }
      
      /* receive */
      if (receiving) {
        for (lSrc = 0; lSrc < GSsrc; ++lSrc) {
          i = lSrc + srcPrc;
          if (sched->rMsgSz[i] > 0) {
	    if (i != proc) {
	      /* Modified by Serge Koren for merge with P++ */
	      PARTI_irecv(NTYPE1+i, RmsgBuf[lSrc], sizeR[lSrc], i, &rGid[lSrc]);
	    }
	  }
	}
      }
      
      if (receiving) {      
	allmesg = 0;
	for (j1 = 0; j1 < grp_size; ++j1)
	  recv_done[j1] = 0;
	while (!allmesg) {
	  allmesg = 1;
	  for (lWait = 0; lWait < GSsrc; ++lWait) {
	    if (!recv_done[lWait]) {
	      i = lWait + srcPrc;
	      if (sched->rMsgSz[i] > 0) {
		if (i != proc) {
                  /* Modified by Serge Koren for merge with P++ */
                  if (PARTI_msgdone(&rGid[lWait])) {
		    recv_done[lWait] = 1;
		  } else {
		    allmesg = 0;
		  }
		} else {
		  recv_done[lWait] = 1;
		}	     
	        if (recv_done[lWait]) {
		  /* unpack receive */
		  recvData = sched->rData[i];
		  destPtr = dest + recvData->startPosn;
		  assert(recvData->numDims <= MAX_DIM);
		  
		  /* set lower bounds for all loops - 0 here */
		  for (currdim = 0; currdim < recvData->numDims; ++currdim) {
		    k[currdim] = 0;
		    tempPtr[currdim] = destPtr;
		  }
		  
		  currdim = done = 0;
		  while (1) {
		    /* reset indices and bump and reset buffer pointers for loops that are done */
		    while (k[currdim] >= recvData->numelem[currdim]) {
		      if (currdim == (recvData->numDims - 1)) {
			done = 1;
			break;
		      }
		      k[currdim] = 0;
		      ++currdim;
		      ++k[currdim];
		      tempPtr[currdim] += recvData->str[currdim];
		      /* reset all inner loop buffer pointers */
		      for (j = currdim-1; j >= 0; --j) {
			tempPtr[j] = tempPtr[j+1];
		      }
		    }
		    
		    if (done)
		      break;

		    /* inner loop code */
		    currdim = 0;
		    *tempPtr[currdim] = *RbufPtr[lWait];
		    ++RbufPtr[lWait];
		    tempPtr[currdim] += recvData->str[currdim];
		    ++k[currdim];
		  }
	        }
	      }
	    }
	  }
	}
      }
      
      if (sending) {
        for (lWait = 0; lWait < GSdest; ++lWait) {
          i = lWait + destPrc;
          if (sched->sMsgSz[i] > 0) {
            if (i != proc) {
	      /* Modified by Serge Koren for merge with P++ */
	      PARTI_msgwait(&sGid[lWait]);
            }
          }
        }
      }
    }
  }
}


void parti_data_move_int(int* array, SCHED* sched, int* dest) {
  int proc, nproc, grp_size, Wprocs;
  int srcPrc, GSsrc, destPrc, GSdest;
  int sending, receiving;
  int i, lPck, lSrc, sizeR[GRPSIZE], sizeS[GRPSIZE];
  int *RbufPtr[GRPSIZE], *SbufPtr[GRPSIZE];
  SchedData *sendData, *recvData;
  int j, k[MAX_DIM], currdim, done;
  int *arrayPtr, *destPtr, *tempPtr[MAX_DIM];
  GID_TYPE sGid[GRPSIZE], rGid[GRPSIZE];
  int j1, allmesg, recv_done[GRPSIZE], lWait;  
  
  if (!msgbuf_done_init)
    init_message_buffers();
 
  if (array == 0 || sched == 0 || dest == 0)
    return;
  
  proc = PARTI_myproc();
  nproc = PARTI_numprocs();
  
  if (GRPSIZE > nproc) {
    grp_size = nproc;
  } else {
    grp_size = GRPSIZE;
  }
  
  Wprocs = (nproc/grp_size)*grp_size;
  
  /* Break 0...(nproc-1) loop into groups of size grp_size */
  for (srcPrc = 0; srcPrc < nproc; srcPrc += grp_size) {
    if (srcPrc >= Wprocs) {
      GSsrc = nproc - Wprocs;
    } else {
      GSsrc = grp_size;
    }    

    GSdest = grp_size;
    
    if ((proc < (srcPrc + GSsrc)) && (proc >= srcPrc)) {
      sending = 1;
    } else {
      sending = 0;
    }
    
    for (destPrc = 0; destPrc < nproc; destPrc += grp_size) {
      if (destPrc >= Wprocs) {
	GSdest = nproc - Wprocs;
      }

      if ((proc < (destPrc + GSdest)) && (proc >= destPrc)) {
	receiving = 1;
      } else {
	receiving = 0;
      }

      /* check receive allocation */
      if (receiving) {
        for (lPck = 0; lPck < GSsrc; ++lPck) {
          i = srcPrc + lPck;
          if (sched->rMsgSz[i] > 0) { 
            sizeR[lPck] = sched->rMsgSz[i]*sizeof(int);	    
            if (i != proc) {
              if (sizeR[lPck] > RmsgBuf_size[lPck]) {
		/* need a bigger message buffer */
		RmsgBuf[lPck] = (char*)realloc(RmsgBuf[lPck], sizeR[lPck]);
                if (RmsgBuf[lPck] == 0) {
                  fprintf(stderr, "Cannot allocate space for message buffers");
                  return;
                }
                RmsgBuf_size[lPck] = sizeR[lPck];
              }
	      RbufPtr[lPck] = (int*)RmsgBuf[lPck];	      
            } else {
              /* copy from the local message buffer */
              if (sizeR[lPck] > local_msgBuf_size) {
                /* need a bigger local message buffer */
		local_msgBuf = (char*)realloc(local_msgBuf, sizeR[lPck]);
                if (local_msgBuf == 0) {
                  fprintf(stderr, "Cannot allocate space for local message buffer");
		  return;
                }
                local_msgBuf_size = sizeR[lPck];
              }
              RbufPtr[lPck] = (int*)local_msgBuf;
	    } 
	  }
	}
      } 
      
      /* check send allocation */
      if (sending) {
        for (lPck = 0; lPck < GSdest; ++lPck) {
	  i = lPck + destPrc;
          if (sched->sMsgSz[i] > 0) {
            sizeS[lPck] = sched->sMsgSz[i]*sizeof(int);
            if (i != proc) {
	      /* prepare a send */
              if (sizeS[lPck] > SmsgBuf_size[lPck]) {
                /* need a bigger message buffer */
		SmsgBuf[lPck] = (char*)realloc(SmsgBuf[lPck], sizeS[lPck]);
                if (SmsgBuf[lPck] == 0) {
                  fprintf(stderr, "Cannot allocate space for message buffers");
                  return;
		}
                SmsgBuf_size[lPck] = sizeS[lPck];
              }
              SbufPtr[lPck] = (int*)SmsgBuf[lPck];
            } else { /* als - 5/19/92 */
	      /* prepare the on-processor copy */
              if (sizeS[lPck] > local_msgBuf_size) {
		/* need a bigger local message buffer */
		local_msgBuf = (char*)realloc(local_msgBuf, sizeS[lPck]);
                if (local_msgBuf == 0) {
                  fprintf(stderr, "Cannot allocate space for local message buffer");
	          return;
                }
                local_msgBuf_size = sizeS[lPck];
              }
              SbufPtr[lPck] = (int*)local_msgBuf;
            }
	    
	    /* pack sends */
            sendData = sched->sData[i];
	    arrayPtr = array + sendData->startPosn;	    
            assert(sendData->numDims <= MAX_DIM);
	    
	    /* set lower bounds for all loops - 0 here */
	    for (currdim = 0; currdim < sendData->numDims; ++currdim) {
	      k[currdim] = 0;
	      tempPtr[currdim] = arrayPtr;
	    }
	    
	    currdim = done = 0;
	    while (1) {
	      /* reset indices and bump and reset buffer pointers for loops that are done */
	      while (k[currdim] >= sendData->numelem[currdim]) {
		if (currdim == (sendData->numDims - 1)) {
		  done = 1;
		  break;
		}		
		k[currdim] = 0;
		++currdim;
		++k[currdim];
		tempPtr[currdim] += sendData->str[currdim];
		/* reset all inner loop buffer pointers */
		for (j = currdim-1; j >= 0; --j) {
		  tempPtr[j] = tempPtr[j+1];
		}
	      }

	      if (done)
		break;

	      /* inner loop code */
	      currdim = 0;
	      *SbufPtr[lPck] = *tempPtr[currdim];
	      ++SbufPtr[lPck];
	      tempPtr[currdim] += sendData->str[currdim];
	      ++k[currdim];
	    }
            
	    if (i != proc) {
	      /* Modified by Serge Koren for merge with P++ */
	      PARTI_isend(NTYPE1+proc, SmsgBuf[lPck], sizeS[lPck], i, &sGid[lPck]);
	    }
	  }
	}
      }
      
      /* receive */
      if (receiving) {
        for (lSrc = 0; lSrc < GSsrc; ++lSrc) {
          i = lSrc + srcPrc;
          if (sched->rMsgSz[i] > 0) {
	    if (i != proc) {
	      /* Modified by Serge Koren for merge with P++ */
	      PARTI_irecv(NTYPE1+i, RmsgBuf[lSrc], sizeR[lSrc], i, &rGid[lSrc]);
	    }
	  }
	}
      }
      
      if (receiving) {      
	allmesg = 0;
	for (j1 = 0; j1 < grp_size; ++j1)
	  recv_done[j1] = 0;
	while (!allmesg) {
	  allmesg = 1;
	  for (lWait = 0; lWait < GSsrc; ++lWait) {
	    if (!recv_done[lWait]) {
	      i = lWait + srcPrc;
	      if (sched->rMsgSz[i] > 0) {
		if (i != proc) {
                  /* Modified by Serge Koren for merge with P++ */
                  if (PARTI_msgdone(&rGid[lWait])) {
		    recv_done[lWait] = 1;
		  } else {
		    allmesg = 0;
		  }
		} else {
		  recv_done[lWait] = 1;
		}	     
	        if (recv_done[lWait]) {
		  /* unpack receive */
		  recvData = sched->rData[i];
		  destPtr = dest + recvData->startPosn;
		  assert(recvData->numDims <= MAX_DIM);
		  
		  /* set lower bounds for all loops - 0 here */
		  for (currdim = 0; currdim < recvData->numDims; ++currdim) {
		    k[currdim] = 0;
		    tempPtr[currdim] = destPtr;
		  }
		  
		  currdim = done = 0;
		  while (1) {
		    /* reset indices and bump and reset buffer pointers for loops that are done */
		    while (k[currdim] >= recvData->numelem[currdim]) {
		      if (currdim == (recvData->numDims - 1)) {
			done = 1;
			break;
		      }
		      k[currdim] = 0;
		      ++currdim;
		      ++k[currdim];
		      tempPtr[currdim] += recvData->str[currdim];
		      /* reset all inner loop buffer pointers */
		      for (j = currdim-1; j >= 0; --j) {
			tempPtr[j] = tempPtr[j+1];
		      }
		    }
		    
		    if (done)
		      break;

		    /* inner loop code */
		    currdim = 0;
		    *tempPtr[currdim] = *RbufPtr[lWait];
		    ++RbufPtr[lWait];
		    tempPtr[currdim] += recvData->str[currdim];
		    ++k[currdim];
		  }
	        }
	      }
	    }
	  }
	}
      }
      
      if (sending) {
        for (lWait = 0; lWait < GSdest; ++lWait) {
          i = lWait + destPrc;
          if (sched->sMsgSz[i] > 0) {
            if (i != proc) {
	      /* Modified by Serge Koren for merge with P++ */
	      PARTI_msgwait(&sGid[lWait]);
            }
          }
        }
      }
    }
  }
}


void parti_data_move_float(float* array, SCHED* sched, float* dest) {
  int proc, nproc, grp_size, Wprocs;
  int srcPrc, GSsrc, destPrc, GSdest;
  int sending, receiving;
  int i, lPck, lSrc, sizeR[GRPSIZE], sizeS[GRPSIZE];
  float *RbufPtr[GRPSIZE], *SbufPtr[GRPSIZE];
  SchedData *sendData, *recvData;
  int j, k[MAX_DIM], currdim, done;
  float *arrayPtr, *destPtr, *tempPtr[MAX_DIM];
  GID_TYPE sGid[GRPSIZE], rGid[GRPSIZE];
  int j1, allmesg, recv_done[GRPSIZE], lWait;  
  
  if (!msgbuf_done_init)
    init_message_buffers();
 
  if (array == 0 || sched == 0 || dest == 0)
    return;
  
  proc = PARTI_myproc();
  nproc = PARTI_numprocs();
  
  if (GRPSIZE > nproc) {
    grp_size = nproc;
  } else {
    grp_size = GRPSIZE;
  }
  
  Wprocs = (nproc/grp_size)*grp_size;
  
  /* Break 0...(nproc-1) loop into groups of size grp_size */
  for (srcPrc = 0; srcPrc < nproc; srcPrc += grp_size) {
    if (srcPrc >= Wprocs) {
      GSsrc = nproc - Wprocs;
    } else {
      GSsrc = grp_size;
    }    

    GSdest = grp_size;
    
    if ((proc < (srcPrc + GSsrc)) && (proc >= srcPrc)) {
      sending = 1;
    } else {
      sending = 0;
    }
    
    for (destPrc = 0; destPrc < nproc; destPrc += grp_size) {
      if (destPrc >= Wprocs) {
	GSdest = nproc - Wprocs;
      }

      if ((proc < (destPrc + GSdest)) && (proc >= destPrc)) {
	receiving = 1;
      } else {
	receiving = 0;
      }

      /* check receive allocation */
      if (receiving) {
        for (lPck = 0; lPck < GSsrc; ++lPck) {
          i = srcPrc + lPck;
          if (sched->rMsgSz[i] > 0) { 
            sizeR[lPck] = sched->rMsgSz[i]*sizeof(float);	    
            if (i != proc) {
              if (sizeR[lPck] > RmsgBuf_size[lPck]) {
		/* need a bigger message buffer */
		RmsgBuf[lPck] = (char*)realloc(RmsgBuf[lPck], sizeR[lPck]);
                if (RmsgBuf[lPck] == 0) {
                  fprintf(stderr, "Cannot allocate space for message buffers");
                  return;
                }
                RmsgBuf_size[lPck] = sizeR[lPck];
              }
	      RbufPtr[lPck] = (float*)RmsgBuf[lPck];	      
            } else {
              /* copy from the local message buffer */
              if (sizeR[lPck] > local_msgBuf_size) {
                /* need a bigger local message buffer */
		local_msgBuf = (char*)realloc(local_msgBuf, sizeR[lPck]);
                if (local_msgBuf == 0) {
                  fprintf(stderr, "Cannot allocate space for local message buffer");
		  return;
                }
                local_msgBuf_size = sizeR[lPck];
              }
              RbufPtr[lPck] = (float*)local_msgBuf;
	    } 
	  }
	}
      } 
      
      /* check send allocation */
      if (sending) {
        for (lPck = 0; lPck < GSdest; ++lPck) {
	  i = lPck + destPrc;
          if (sched->sMsgSz[i] > 0) {
            sizeS[lPck] = sched->sMsgSz[i]*sizeof(float);
            if (i != proc) {
	      /* prepare a send */
              if (sizeS[lPck] > SmsgBuf_size[lPck]) {
                /* need a bigger message buffer */
		SmsgBuf[lPck] = (char*)realloc(SmsgBuf[lPck], sizeS[lPck]);
                if (SmsgBuf[lPck] == 0) {
                  fprintf(stderr, "Cannot allocate space for message buffers");
                  return;
		}
                SmsgBuf_size[lPck] = sizeS[lPck];
              }
              SbufPtr[lPck] = (float*)SmsgBuf[lPck];
            } else { /* als - 5/19/92 */
	      /* prepare the on-processor copy */
              if (sizeS[lPck] > local_msgBuf_size) {
		/* need a bigger local message buffer */
		local_msgBuf = (char*)realloc(local_msgBuf, sizeS[lPck]);
                if (local_msgBuf == 0) {
                  fprintf(stderr, "Cannot allocate space for local message buffer");
	          return;
                }
                local_msgBuf_size = sizeS[lPck];
              }
              SbufPtr[lPck] = (float*)local_msgBuf;
            }
	    
	    /* pack sends */
            sendData = sched->sData[i];
	    arrayPtr = array + sendData->startPosn;	    
            assert(sendData->numDims <= MAX_DIM);
	    
	    /* set lower bounds for all loops - 0 here */
	    for (currdim = 0; currdim < sendData->numDims; ++currdim) {
	      k[currdim] = 0;
	      tempPtr[currdim] = arrayPtr;
	    }
	    
	    currdim = done = 0;
	    while (1) {
	      /* reset indices and bump and reset buffer pointers for loops that are done */
	      while (k[currdim] >= sendData->numelem[currdim]) {
		if (currdim == (sendData->numDims - 1)) {
		  done = 1;
		  break;
		}		
		k[currdim] = 0;
		++currdim;
		++k[currdim];
		tempPtr[currdim] += sendData->str[currdim];
		/* reset all inner loop buffer pointers */
		for (j = currdim-1; j >= 0; --j) {
		  tempPtr[j] = tempPtr[j+1];
		}
	      }

	      if (done)
		break;

	      /* inner loop code */
	      currdim = 0;
	      *SbufPtr[lPck] = *tempPtr[currdim];
	      ++SbufPtr[lPck];
	      tempPtr[currdim] += sendData->str[currdim];
	      ++k[currdim];
	    }
            
	    if (i != proc) {
	      /* Modified by Serge Koren for merge with P++ */
	      PARTI_isend(NTYPE1+proc, SmsgBuf[lPck], sizeS[lPck], i, &sGid[lPck]);
	    }
	  }
	}
      }
      
      /* receive */
      if (receiving) {
        for (lSrc = 0; lSrc < GSsrc; ++lSrc) {
          i = lSrc + srcPrc;
          if (sched->rMsgSz[i] > 0) {
	    if (i != proc) {
	      /* Modified by Serge Koren for merge with P++ */
	      PARTI_irecv(NTYPE1+i, RmsgBuf[lSrc], sizeR[lSrc], i, &rGid[lSrc]);
	    }
	  }
	}
      }
      
      if (receiving) {      
	allmesg = 0;
	for (j1 = 0; j1 < grp_size; ++j1)
	  recv_done[j1] = 0;
	while (!allmesg) {
	  allmesg = 1;
	  for (lWait = 0; lWait < GSsrc; ++lWait) {
	    if (!recv_done[lWait]) {
	      i = lWait + srcPrc;
	      if (sched->rMsgSz[i] > 0) {
		if (i != proc) {
                  /* Modified by Serge Koren for merge with P++ */
                  if (PARTI_msgdone(&rGid[lWait])) {
		    recv_done[lWait] = 1;
		  } else {
		    allmesg = 0;
		  }
		} else {
		  recv_done[lWait] = 1;
		}	     
	        if (recv_done[lWait]) {
		  /* unpack receive */
		  recvData = sched->rData[i];
		  destPtr = dest + recvData->startPosn;
		  assert(recvData->numDims <= MAX_DIM);
		  
		  /* set lower bounds for all loops - 0 here */
		  for (currdim = 0; currdim < recvData->numDims; ++currdim) {
		    k[currdim] = 0;
		    tempPtr[currdim] = destPtr;
		  }
		  
		  currdim = done = 0;
		  while (1) {
		    /* reset indices and bump and reset buffer pointers for loops that are done */
		    while (k[currdim] >= recvData->numelem[currdim]) {
		      if (currdim == (recvData->numDims - 1)) {
			done = 1;
			break;
		      }
		      k[currdim] = 0;
		      ++currdim;
		      ++k[currdim];
		      tempPtr[currdim] += recvData->str[currdim];
		      /* reset all inner loop buffer pointers */
		      for (j = currdim-1; j >= 0; --j) {
			tempPtr[j] = tempPtr[j+1];
		      }
		    }
		    
		    if (done)
		      break;

		    /* inner loop code */
		    currdim = 0;
		    *tempPtr[currdim] = *RbufPtr[lWait];
		    ++RbufPtr[lWait];
		    tempPtr[currdim] += recvData->str[currdim];
		    ++k[currdim];
		  }
	        }
	      }
	    }
	  }
	}
      }
      
      if (sending) {
        for (lWait = 0; lWait < GSdest; ++lWait) {
          i = lWait + destPrc;
          if (sched->sMsgSz[i] > 0) {
            if (i != proc) {
	      /* Modified by Serge Koren for merge with P++ */
	      PARTI_msgwait(&sGid[lWait]);
            }
          }
        }
      }
    }
  }
}


void parti_data_move_double(double* array, SCHED* sched, double* dest) {
  int proc, nproc, grp_size, Wprocs;
  int srcPrc, GSsrc, destPrc, GSdest;
  int sending, receiving;
  int i, lPck, lSrc, sizeR[GRPSIZE], sizeS[GRPSIZE];
  double *RbufPtr[GRPSIZE], *SbufPtr[GRPSIZE];
  SchedData *sendData, *recvData;
  int j, k[MAX_DIM], currdim, done;
  double *arrayPtr, *destPtr, *tempPtr[MAX_DIM];
  GID_TYPE sGid[GRPSIZE], rGid[GRPSIZE];
  int j1, allmesg, recv_done[GRPSIZE], lWait;  
  
  if (!msgbuf_done_init)
    init_message_buffers();
 
  if (array == 0 || sched == 0 || dest == 0)
    return;
  
  proc = PARTI_myproc();
  nproc = PARTI_numprocs();
  
  if (GRPSIZE > nproc) {
    grp_size = nproc;
  } else {
    grp_size = GRPSIZE;
  }
  
  Wprocs = (nproc/grp_size)*grp_size;
  
  /* Break 0...(nproc-1) loop into groups of size grp_size */
  for (srcPrc = 0; srcPrc < nproc; srcPrc += grp_size) {
    if (srcPrc >= Wprocs) {
      GSsrc = nproc - Wprocs;
    } else {
      GSsrc = grp_size;
    }    

    GSdest = grp_size;
    
    if ((proc < (srcPrc + GSsrc)) && (proc >= srcPrc)) {
      sending = 1;
    } else {
      sending = 0;
    }
    
    for (destPrc = 0; destPrc < nproc; destPrc += grp_size) {
      if (destPrc >= Wprocs) {
	GSdest = nproc - Wprocs;
      }

      if ((proc < (destPrc + GSdest)) && (proc >= destPrc)) {
	receiving = 1;
      } else {
	receiving = 0;
      }

      /* check receive allocation */
      if (receiving) {
        for (lPck = 0; lPck < GSsrc; ++lPck) {
          i = srcPrc + lPck;
          if (sched->rMsgSz[i] > 0) { 
            sizeR[lPck] = sched->rMsgSz[i]*sizeof(double);	    
            if (i != proc) {
              if (sizeR[lPck] > RmsgBuf_size[lPck]) {
		/* need a bigger message buffer */
		RmsgBuf[lPck] = (char*)realloc(RmsgBuf[lPck], sizeR[lPck]);
                if (RmsgBuf[lPck] == 0) {
                  fprintf(stderr, "Cannot allocate space for message buffers");
                  return;
                }
                RmsgBuf_size[lPck] = sizeR[lPck];
              }
	      RbufPtr[lPck] = (double*)RmsgBuf[lPck];	      
            } else {
              /* copy from the local message buffer */
              if (sizeR[lPck] > local_msgBuf_size) {
                /* need a bigger local message buffer */
		local_msgBuf = (char*)realloc(local_msgBuf, sizeR[lPck]);
                if (local_msgBuf == 0) {
                  fprintf(stderr, "Cannot allocate space for local message buffer");
		  return;
                }
                local_msgBuf_size = sizeR[lPck];
              }
              RbufPtr[lPck] = (double*)local_msgBuf;
	    } 
	  }
	}
      } 
      
      /* check send allocation */
      if (sending) {
        for (lPck = 0; lPck < GSdest; ++lPck) {
	  i = lPck + destPrc;
          if (sched->sMsgSz[i] > 0) {
            sizeS[lPck] = sched->sMsgSz[i]*sizeof(double);
            if (i != proc) {
	      /* prepare a send */
              if (sizeS[lPck] > SmsgBuf_size[lPck]) {
                /* need a bigger message buffer */
		SmsgBuf[lPck] = (char*)realloc(SmsgBuf[lPck], sizeS[lPck]);
                if (SmsgBuf[lPck] == 0) {
                  fprintf(stderr, "Cannot allocate space for message buffers");
                  return;
		}
                SmsgBuf_size[lPck] = sizeS[lPck];
              }
              SbufPtr[lPck] = (double*)SmsgBuf[lPck];
            } else { /* als - 5/19/92 */
	      /* prepare the on-processor copy */
              if (sizeS[lPck] > local_msgBuf_size) {
		/* need a bigger local message buffer */
		local_msgBuf = (char*)realloc(local_msgBuf, sizeS[lPck]);
                if (local_msgBuf == 0) {
                  fprintf(stderr, "Cannot allocate space for local message buffer");
	          return;
                }
                local_msgBuf_size = sizeS[lPck];
              }
              SbufPtr[lPck] = (double*)local_msgBuf;
            }
	    
	    /* pack sends */
            sendData = sched->sData[i];
	    arrayPtr = array + sendData->startPosn;	    
            assert(sendData->numDims <= MAX_DIM);
	    
	    /* set lower bounds for all loops - 0 here */
	    for (currdim = 0; currdim < sendData->numDims; ++currdim) {
	      k[currdim] = 0;
	      tempPtr[currdim] = arrayPtr;
	    }
	    
	    currdim = done = 0;
	    while (1) {
	      /* reset indices and bump and reset buffer pointers for loops that are done */
	      while (k[currdim] >= sendData->numelem[currdim]) {
		if (currdim == (sendData->numDims - 1)) {
		  done = 1;
		  break;
		}		
		k[currdim] = 0;
		++currdim;
		++k[currdim];
		tempPtr[currdim] += sendData->str[currdim];
		/* reset all inner loop buffer pointers */
		for (j = currdim-1; j >= 0; --j) {
		  tempPtr[j] = tempPtr[j+1];
		}
	      }

	      if (done)
		break;

	      /* inner loop code */
	      currdim = 0;
	      *SbufPtr[lPck] = *tempPtr[currdim];
	      ++SbufPtr[lPck];
	      tempPtr[currdim] += sendData->str[currdim];
	      ++k[currdim];
	    }
            
	    if (i != proc) {
	      /* Modified by Serge Koren for merge with P++ */
	      PARTI_isend(NTYPE1+proc, SmsgBuf[lPck], sizeS[lPck], i, &sGid[lPck]);
	    }
	  }
	}
      }
      
      /* receive */
      if (receiving) {
        for (lSrc = 0; lSrc < GSsrc; ++lSrc) {
          i = lSrc + srcPrc;
          if (sched->rMsgSz[i] > 0) {
	    if (i != proc) {
	      /* Modified by Serge Koren for merge with P++ */
	      PARTI_irecv(NTYPE1+i, RmsgBuf[lSrc], sizeR[lSrc], i, &rGid[lSrc]);
	    }
	  }
	}
      }
      
      if (receiving) {      
	allmesg = 0;
	for (j1 = 0; j1 < grp_size; ++j1)
	  recv_done[j1] = 0;
	while (!allmesg) {
	  allmesg = 1;
	  for (lWait = 0; lWait < GSsrc; ++lWait) {
	    if (!recv_done[lWait]) {
	      i = lWait + srcPrc;
	      if (sched->rMsgSz[i] > 0) {
		if (i != proc) {
                  /* Modified by Serge Koren for merge with P++ */
                  if (PARTI_msgdone(&rGid[lWait])) {
		    recv_done[lWait] = 1;
		  } else {
		    allmesg = 0;
		  }
		} else {
		  recv_done[lWait] = 1;
		}	     
	        if (recv_done[lWait]) {
		  /* unpack receive */
		  recvData = sched->rData[i];
		  destPtr = dest + recvData->startPosn;
		  assert(recvData->numDims <= MAX_DIM);
		  
		  /* set lower bounds for all loops - 0 here */
		  for (currdim = 0; currdim < recvData->numDims; ++currdim) {
		    k[currdim] = 0;
		    tempPtr[currdim] = destPtr;
		  }
		  
		  currdim = done = 0;
		  while (1) {
		    /* reset indices and bump and reset buffer pointers for loops that are done */
		    while (k[currdim] >= recvData->numelem[currdim]) {
		      if (currdim == (recvData->numDims - 1)) {
			done = 1;
			break;
		      }
		      k[currdim] = 0;
		      ++currdim;
		      ++k[currdim];
		      tempPtr[currdim] += recvData->str[currdim];
		      /* reset all inner loop buffer pointers */
		      for (j = currdim-1; j >= 0; --j) {
			tempPtr[j] = tempPtr[j+1];
		      }
		    }
		    
		    if (done)
		      break;

		    /* inner loop code */
		    currdim = 0;
		    *tempPtr[currdim] = *RbufPtr[lWait];
		    ++RbufPtr[lWait];
		    tempPtr[currdim] += recvData->str[currdim];
		    ++k[currdim];
		  }
	        }
	      }
	    }
	  }
	}
      }
      
      if (sending) {
        for (lWait = 0; lWait < GSdest; ++lWait) {
          i = lWait + destPrc;
          if (sched->sMsgSz[i] > 0) {
            if (i != proc) {
	      /* Modified by Serge Koren for merge with P++ */
	      PARTI_msgwait(&sGid[lWait]);
            }
          }
        }
      }
    }
  }
}

