/*
 * Auth: Alan Sussman (als@cs.umd.edu)
 *       University of Maryland, College Park
 * Desc: These functions are depreciated in favor
 *       of the irregular block decomposition code
 */

#include "decomp.h"
#include "utils.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#define min(A,B) ((A)<(B)?(A):(B))
#define max(A,B) ((A)>(B)?(A):(B))

/* DISABLED THE BODY OF THIS FUNCTION - als and gupta - 2/7/92 */
int getDecompDim(DARRAY* dArray, int dim) {
  return dim;
}

/***************************************************************************
 *
 * vProc( numDims, sizes )
 * 
 * Inputs: vProcPtr  - 
 *         numDims   - the number of dimensions in virtual processor space
 *                     LIMITATION : Only 1 DIM
 *         sizes  - array of the number of virtual processors in
 *                  each dimension
 *
 * Creates virtual processor space
 *
 * Virtual processor numbers start at 0 in each dimension (For C)
 **************************************************************************/
VPROC* vProc(int numDims, int* sizes) {
  int i, nprocs;
  VPROC *vProcPtr;
  
  nprocs = PARTI_numprocs();
  
  /* for now assume vproc has only 1 dimension */
  vProcPtr = (VPROC*)pmalloc("VPROC", sizeof(VPROC));
  vProcPtr->nProcs = nprocs;
  vProcPtr->usedProcs = 0;  
  vProcPtr->nDims = numDims;
  vProcPtr->dimVec = (int*)calloc(numDims, sizeof(int));
  for (i = 0; i < numDims; i++) {
    vProcPtr->dimVec[i] = sizes[i];
  }
  
  return vProcPtr;
}

/*****************************************************************************
 *
 *  create_decomp(numDims, sizes)
 * 
 *  creates the a decomposition with "numDim" dimensions whose sizes are
 *  given by the array sizes
 * 
 * Inputs: numDims - the dimension of dArray whose overlap/ghost cells are 
 *                   being updated
 *         sizes -  the sizes of the decomposition in each dimension
 *                 
 * 
 * Returns: pointer to the newly created decomposition
 * 
 *****************************************************************************/
struct decomp_rec* xcreate_decomp(int numDims, int* sizes) {
  struct decomp_rec *decomp;
  int i;
  
  decomp = (struct decomp_rec*)pmalloc("struct decomp_rec", sizeof(struct decomp_rec));  
  decomp->nDims = numDims;
  decomp->dimVec = (int*)calloc(decomp->nDims, sizeof(int));
  decomp->dimProc = (int*)calloc(decomp->nDims, sizeof(int));
  decomp->dimDist = (char*)calloc(decomp->nDims, sizeof(char));
  decomp->baseProc = 0;
  decomp->nProcs = PARTI_numprocs();
  decomp->referenceCount = 0;
  
  for (i = 0; i < numDims; i++) {
    decomp->dimVec[i] = sizes[i];
  }
  
  return decomp;
}

DECOMP* create_decomp(int numDims, int* sizes) {
  parti_decomp_t* decomp;
  decomp = parti_create_idecomp(numDims, sizes);
  decomp->temp = xcreate_decomp(numDims, sizes);
  return decomp;
}

/******************************************************************************
 *  distribute(decomp, dist)
 * 
 *  set the type of distribution for each dimension of the decomposition
 * 
 * Inputs:   decomp    pointer to the decomposition
 *           dist      string of N character where (N = numDim).
 *                     Each character can have one of the following 3 values
 *                      'B' Blocked distribution
 *                      'C' Cyclic distribution
 *                      '*' No distribution
 *                 
 * 
 * Returns: pointer to the newly created decomposition
 * 
 ***************************************************************************/
void xdistribute(struct decomp_rec* decomp, char* dist) {
  int i, used, bigRatioDim, nProcs, factor;
  float ratio, nRatio;

  for (i = 0; i < decomp->nDims; i++) {
    decomp->dimDist[i] = dist[i];  
  /* decide the number of processors allocated to each dimension of the decomposition. */
    decomp->dimProc[i] = 1;
  }
 
  /* the number of physical processors used by the decomposition */
  nProcs = decomp->nProcs;
  
  /* don't know if this shaping of the decomposition (from a linear
     processor array into the number of dimensions of the decomposition)
     always produced the properly shaped decomposition     
     changed to use greatest_prime_factor() 2/6/92 - als
  */  
  used = 1;
  while (used < nProcs) {
    bigRatioDim = 0;
    ratio = 1;
    if (decomp->dimDist[0] != '*')
      ratio = ((float)decomp->dimVec[0])/((float)decomp->dimProc[0]);
    for (i = 1; i < decomp->nDims; i++) {
      if (decomp->dimDist[i] != '*')
	nRatio = ((float)decomp->dimVec[i])/((float)decomp->dimProc[i]);
      else
	nRatio = 1;
      if (nRatio > ratio || (nRatio == ratio && nRatio > 1 && decomp->dimProc[i] < decomp->dimProc[bigRatioDim])) {
	/* if the ratios are the same, this favors the dimension that */
	/* already is assigned fewer processors */
	ratio = nRatio;
	bigRatioDim = i;
      }
    }
    factor = greatest_prime_factor(nProcs/used);
    decomp->dimProc[bigRatioDim] *= factor;
    used *= factor;
  }
}

void distribute(DECOMP* decomp, char* dist) {
  xdistribute(decomp->temp, dist);
}

/***************************************************************************
 *
 * embed(decomp, vproc, startPosn, endPosn) 
 * 
 *  embed a decomposition in the virtual processor space.  This involves
 *  deciding what processors are allocated to the decomposition.
 *  
 * 
 * Inputs: decomp  - pointer to the decomposition
 *         vproc   - pointer to the virtual processor space
 *         startPosn - start position in the virtual processor space
 *         endPosn   - end position in the virtual processor sapce
 * 
 *
 ***************************************************************************/
void xembed(struct decomp_rec* decomp, VPROC* vp, int startPosn, int endPosn) {
  int firstProc, lastProc;
  float temp;
  
  /* 
   * decide what physical processors are allocated to a decomposition
   * temp is the number of virtual processors per physical processor 
   */
  temp = (float)vp->dimVec[0] / vp->nProcs;
  firstProc = (int)startPosn / temp;
  lastProc = max((int)(endPosn/temp), 0);
  
  /* 
   * if both physical and virtual processors are numbered starting at 0,
   * then this works
   * this doesn't work if number of v.p.'s is not evenly divisible by the
   * number of physical processors
   * lastProc  = endPosn/temp; 
   */  
  if ((firstProc < 0) || (lastProc >= vp->nProcs)) {
    fprintf(stderr, "Attempt to allocate to an illegal processor");
    return;
  }
  decomp->nProcs = lastProc - firstProc + 1;
  decomp->baseProc = firstProc;
  vp->usedProcs += decomp->nProcs; 
}

void embed(DECOMP* decomp, VPROC* vp, int startPosn, int endPosn) {
  xembed(decomp->temp, vp, startPosn, endPosn);
}

int section(DECOMP* decomp, int* dim, int* cnt, int** idx) {
  return parti_section(decomp, dim, cnt, idx);
}

int partition(DECOMP* decomp, int dim, int* blk, int cnt, int* idx) {
  return parti_partition(decomp, dim, blk, cnt, idx);
}

/* Rewritten as per the new definition  of Sched : Gagan   07/03/93   */ 
void init_Sched(SCHED* sched) {
  static int ID_Counter = 1;

  if (sched == NULL)
    return;
  
  memset(sched->sMsgSz, 0, MAX_NODES*sizeof(int));
  memset(sched->rMsgSz, 0, MAX_NODES*sizeof(int));
  memset(sched->rData, 0, MAX_NODES*sizeof(SchedData*));
  memset(sched->sData, 0, MAX_NODES*sizeof(SchedData*));
  sched->type = 0;
  sched->hash = -1;
  sched->referenceCount = 0;
  sched->ID_Number = ID_Counter++;

#if PARTI_DEBUG > 2
  printf("In init_Sched: X->ID_Number = %d\n", sched->ID_Number);
#endif
}
