/*
 * Auth: Alan Sussman (als@cs.umd.edu)
 *       University of Maryland, College Park
 * Desc:
 */

#include "helper.h"
#include <assert.h>

#define min(A,B) ((A)<(B)?(A):(B))
#define max(A,B) ((A)>(B)?(A):(B))

/**************************************************************************
 *
 * if node is one of the processors assigned to the decomposition decomp 
 * then return TRUE otherwise FALSE
 *
 **************************************************************************/
int active(int node, struct decomp_rec* decomp) {
  int baseProc, maxProc;
  
  baseProc = decomp->baseProc;
  maxProc = baseProc + decomp->nProcs;
  
  return ((node >= baseProc) && (node < maxProc));
}

/* gray should be called from distribute(), not align(), since it works
   on a decomposition, not a darray */
/*
  Gray computes the coordinates of physical processor p in decomposition
  vMach, returning them in posVec - gray assumes that the processors are
  linearly numbered, with the decomposition embedded starting at processor
  vMach -> baseProc
*/
void gray(struct decomp_rec* vMach, int p, int* posVec) {
  int i, tmp = 1;

  if ((p < vMach->baseProc) || (p >= vMach->baseProc+vMach->nProcs)) {
    for (i = 0; i < vMach->nDims; i++)
      posVec[i] = -1;
    return;
  }
  
  p -= vMach->baseProc;
  /* tmp becomes the total number of physical processors used by
     the decomposition */
  for (i = 1; i < vMach->nDims; i++)
    tmp *= vMach->dimProc[i];
  
  assert (tmp > 0);
  
  for (i = 0; i < vMach->nDims - 1; i++) {
    posVec[i] = max(0, p/tmp);
    p -= posVec[i]*tmp;
    tmp /= vMach->dimProc[i+1];
  }
  posVec[i] = max(0, p);
}

int invGray(struct decomp_rec* vMach, int* posVec) {
  int ndims, *szVec, i, p = 0, tmp = 1;
  
  ndims = vMach->nDims;
  szVec = vMach->dimProc;
  for (i = 1; i <= ndims; i++) {
    if ((posVec[ndims-i] < 0) || (posVec[ndims-i] >= szVec[ndims-i])) {
      return -1;
    }
    p += posVec[ndims-i] * tmp;
    tmp *= szVec[ndims-i];
  }

  p += vMach->baseProc;

  if ((p < vMach->baseProc) || (p >= vMach->baseProc + vMach->nProcs)) {
    return -1;
  } else {
    return p;
  }
}
