/*
 * Auth: Alan Sussman (als@cs.umd.edu)
 *       University of Maryland, College Park
 * Desc:
 */

#ifndef LIST_H
#define LIST_H

#include "bsparti.h"

typedef void gtype;           /* generic type */
typedef int (*pfunc)();       /* pointer to function (for EQ functions) */

typedef struct _thing_cell {  /* a link in a list of the things in the proc */
  gtype* item;
  struct _thing_cell* next;
  struct _thing_cell* prev;
  int referenceCount;         /* Code added by Dan Quinlan */
} cell;

typedef struct _list {
  cell* dummy;                /* ptr to the dummy cell */
  cell* fptr;                 /* scratch ptr used by FOREACH */
  cell* optr;                 /* scratch ptr used by NEXT_ITEM */
  int referenceCount;         /* Code added by Dan Quinlan */
}* list;

list create_list();                          /* create an empty list */
void destroy_list(list);                     /* destroy a list */
void insert_list(gtype*, list);              /* insert an element into a list */
gtype* remove_list(gtype*, list);            /* remove an item from a list */
void remove_list_list(list, list);           /* remove elements of l1 from list l2 */
gtype* member_list(gtype*, list);            /* membership test for lists */
gtype* eq_member_list(gtype*, list, pfunc);  /* membership with EQ func */
int length_list(list);                       /* number of elements in a list */
list dup_list(list);                         /* duplicate a list */
void append_list(list, list);                /* l1 <-- l1 + l2 */      
list intersection_list(list, list);          /* intersection of two lists */
list union_list(list, list);                 /* union of two lists */
void replace_list(gtype*,gtype*,list);       /* change an item in a list */
/*void delete_cell(cell*);*/                     /* Added by Dan Quinlan */
void behead_list(list);

/*
 * These macros can be used as follows
 *    cell *thing;
 *    list l;
 *       for (thing = FIRST(l); STILL_IN(thing, l); thing = NEXT(thing)) {
 *          ...
 *       }
 *    }
 *
 *   gtype* item;
 *   FOREACH(item, l) { ...just like a for loop...  }
 *
 *   item = FIRST_ITEM(l)
 *   while (item) { ... just like a while loop...;  item = NEXT_ITEM(l); }
 *
 *   item = LAST_ITEM(l)
 *   while (item) { ... backwards too! ...; item = PREV_ITEM(l); }
 */
#define FIRST(l)           (l->dummy->next)
#define FIRST_ITEM(l)      (l->optr = FIRST(l), l->optr->item)
#define LAST(l)            (l->dummy->prev)
#define LAST_ITEM(l)       (l->optr = LAST(l), l->optr->item)
#define EMPTY(l)           (l->dummy == FIRST(l))
#define STILL_IN(t, l)     (t != l->dummy)
#define NEXT(l)            (l->next)
#define NEXT_ITEM(l)       (l->optr = NEXT(l->optr), l->optr->item)
#define PREV(l)            (l->prev)
#define PREV_ITEM(l)       (l->optr = PREV(l->optr), l->optr->item)
#define FOREACH(i,l)       for(l->fptr = FIRST(l), i = l->fptr->item; \
                             STILL_IN(l->fptr, l); \
                             l->fptr = NEXT(l->fptr), i = l->fptr->item)
#define FOREACHt(i,l,type) for(l->fptr = FIRST(l), i = (type) l->fptr->item; \
                             STILL_IN(l->fptr, l); \
                             l->fptr = NEXT(l->fptr), i = (type) l->fptr->item)

#endif /* LIST_H */
