/*
 * Auth: Alan Sussman (als@cs.umd.edu)
 *       University of Maryland, College Park
 * Desc:
 */

#include "print.h"
#include <stdlib.h>
#include <stdio.h>

void print_schedData(SchedData* schedData) {
  int i, ndims, startPosn, *step, *nelems;
  
  printf("----------------------------\n");
  ndims = schedData->numDims;
  startPosn = schedData->startPosn;
  nelems = schedData->numelem;
  step = schedData->str;
  printf("Communicating %d dimensional array, starting at position %d \n",
	 ndims, startPosn);
  for (i = 0; i < ndims; i++) {
    printf("Dim. %d: %d total elements at stride %d apart \n", i, nelems[i], step[i]);
  }
}

void print_sched(SCHED* schedPtr) {
  int i, j, numprocs, proc;
  
  proc = PARTI_myproc();
  numprocs = PARTI_numprocs();
  
  if (schedPtr == NULL) {
    printf("Node %d has empty schedule \n", proc);
    return;
  }
  
  for (j = 0; j < numprocs; j++) {
    PARTI_gsync();
    if (j == proc) {
      for (i = 0; i < numprocs; i++) {
	if (schedPtr->sMsgSz[i] > 0) {
	  
	  printf("NODE %d : send schedule to proc %d of %d elements \n", 
		 proc, i, schedPtr->sMsgSz[i]);
	  
	  print_schedData(schedPtr->sData[i]);  
	}
      }
      for (i = 0; i < numprocs; i++) {
	if (schedPtr->rMsgSz[i] > 0) {
	  
	  printf("NODE %d : receive schedule from proc %d of %d elements \n", 
		 proc, i, schedPtr->rMsgSz[i]);
	  
	  print_schedData(schedPtr->rData[i]); 
	}
      }
      printf("\n");
    }
  }
}

void print_msgBuf(int* msgBuf, int len) {
  int i;
  
  printf("\nNODE %d: msgBuf = ",PARTI_myproc());
  for (i = 0; i < len; i++) {
    if (i % 10 == 0) printf("\n");
    printf("%d ", msgBuf[i]);
  }
  printf("\n");
}
