/*
 * Auth: Alan Sussman (als@cs.umd.edu)
 *       University of Maryland, College Park
 * Desc:
 */

#include "port.h"
#include "hash.h"
#include "config.h"
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

#ifdef USE_PVM
static char* parti_pvm_group = 0;

int PARTI_PVM_setup(char* grp) {
  assert(grp != NULL);
#if PARTI_DEBUG > 3
  fprintf(stderr, "PARTI_PVM_setup(%s)\n", grp);
#endif
  parti_pvm_group = strdup(grp);
  assert(PARTI_PVM_numprocs() < MAX_NODES);
  return 0;
}

void PARTI_PVM_cleanup() {
#if PARTI_DEBUG > 3
  fprintf(stderr, "PARTI_PVM_cleanup()\n");
#endif
  cleanup_after_PARTI(); /* destroy hash tables */
  if (parti_pvm_group) 
    free(parti_pvm_group);
}

int PARTI_PVM_numprocs() {
  int rv;
  
#if PARTI_DEBUG > 3
  fprintf(stderr, "PARTI_PVM_numprocs(): ");
#endif
  if ((rv = pvm_gsize(parti_pvm_group)) < 0) {
    pvm_perror("gsize");
  }
#if PARTI_DEBUG > 3
  fprintf(stderr, " returning %d\n", rv);
#endif
  return rv;
}

int PARTI_PVM_gsync() {
#if PARTI_DEBUG > 3
  fprintf(stderr, "PARTI_PVM_gsync(): calling pvm_barrier(%s)\n", parti_pvm_group);
#endif
  return pvm_barrier(parti_pvm_group, pvm_gsize(parti_pvm_group));
}

int PARTI_PVM_myproc() {
  int rv;
#if PARTI_DEBUG > 3
  fprintf(stderr, "PARTI_PVM_myproc(): ");
#endif
  if ((rv = pvm_getinst(parti_pvm_group, pvm_mytid())) < 0)
    pvm_perror("getinst");
#if PARTI_DEBUG > 3
  fprintf(stderr, "returning %d\n", rv);
#endif
  return rv;
}

void PARTI_PVM_killall() {
#if PARTI_DEBUG > 3
  fprintf(stderr, "PARTI_PVM_killall()\n");
#endif
  if(pvm_lvgroup(parti_pvm_group) < 0)
    pvm_perror("lvgroup");
  pvm_exit();
}

int PARTI_PVM_csend(int tag, char* buf, int len, int node) {
#if PARTI_DEBUG > 3
  fprintf(stderr, "PARTI_PVM_csend(%d, %d, %d)\n", tag, len, node);
#endif
  pvm_initsend(PvmDataDefault);
  pvm_pkbyte(buf, len, 1);
  pvm_send(pvm_gettid(parti_pvm_group, node), tag);
  
  return 0;
}

int PARTI_PVM_crecv(int tag, char* buf, int len, int node) {
#if PARTI_DEBUG > 3
  fprintf(stderr, "PARTI_PVM_crecv(%d, %d, %d)\n", tag, len, node);
#endif
  pvm_recv(pvm_gettid(parti_pvm_group, node), tag);
  pvm_upkbyte(buf, len, 1);
  
  return 0;
}

#ifdef FORTRAN

#define PARTISETUP_F77   F77_FUNC(partisetup, PARTISETUP)
#define PARTICLEANUP_F77 F77_FUNC(particleanup, PARTICLEANUP)

/* added for fortran */
void PARTISETUP_F77(char* grp, int len) {
  int i;
  char* tmp;

  tmp = (char*)calloc(len+1, sizeof(char));
  for (i = 0; i < len; ++i) {
    tmp[i] = grp[i];
  }
  tmp[i] = 0;
  (void)PARTI_PVM_setup(tmp);
  free(tmp);
}

/* added for fortran */
void PARTICLEANUP_F77() {
  /* destroy all the fortran objects */
  destroy_sch_table();
  destroy_da_table();
  destroy_decomp_table();
  destroy_vproc_table();
  /* destroy all the object caches */
  PARTI_PVM_cleanup();
}

#endif /* USE_PVM */

#endif /* FORTRAN */
