/*
 * Auth: Alan Sussman (als@cs.umd.edu)
 *       University of Maryland, College Park
 * Desc:
 */

#include "bsparti.h"
#include "decomp.h"
#include "hash.h"
#include "helper.h"
#include "utils.h"
#include <stdlib.h>
#include <stdio.h>

SCHED* subArraySched(DARRAY* srcDA, DARRAY* destDA, int numDims,
		     int* srcDims, int* sLos, int* sHis, int* sStrides,
		     int* destDims, int* dLos, int* dHis, int* dStrides) {
  parti_sched_t* sched;
  parti_region_t* srceRegion, *destRegion;
  int* map, i;

  srceRegion = parti_alloc_region(numDims);
  destRegion = parti_alloc_region(numDims);
  map = (int*)calloc(numDims, sizeof(int));
  
  for (i = 0; i < numDims; ++i) {
    srceRegion->lower[srcDims[i]] = sLos[srcDims[i]];
    srceRegion->upper[srcDims[i]] = sHis[srcDims[i]];
    srceRegion->stride[srcDims[i]] = sStrides[srcDims[i]];
    destRegion->lower[destDims[i]] = dLos[destDims[i]];
    destRegion->upper[destDims[i]] = dHis[destDims[i]];
    destRegion->stride[destDims[i]] = dStrides[destDims[i]];
    map[srcDims[i]] = destDims[i];
  }
  
  sched = parti_region_sched(srcDA, srceRegion, destDA, destRegion, map);

  free(map);
  parti_free_region(destRegion);
  parti_free_region(srceRegion);
  
  return sched;
}
