/* 
 * Auth: Christian Hansen (chansen@cs.umd.edu)
 *       University of Maryland, College Park
 * Desc: Initializes MBP to use MPI
 */

#include <mpi.h>
#include <multiblockparti.h>
#include <stdio.h>
#include <string.h>

int main(int argc, char* argv[]) {
  int p, r;
  char *f, fn[80];
  FILE* ifp, *ofp;
  MPI_Comm groupcomm;
  
  /* initialize MPI */
  MPI_Init(&argc, &argv);
  MPI_Comm_dup(MPI_COMM_WORLD, &groupcomm);
  MPI_Comm_size(groupcomm, &p);
  MPI_Barrier(groupcomm);
  
  /* check command-line args */
  if (argc > 1) {
    if (!(ifp = fopen(argv[1], "r"))) {
      fprintf(stderr, "could not open test case file\n");
      return 1;
    }
  } else {
    ifp = stdin;
  }

  /* initialize MBP */
  parti_setup(&groupcomm);
  
  /* run tests */
  MPI_Comm_rank(groupcomm, &r);  
  if (!(f = (char*)strrchr(argv[0], '/')))
    f = argv[0];
  sprintf(fn, "/tmp/%s.out.%d", ++f, r);
  if (ofp = fopen(fn, "w")) { 
    parti_test(ifp, ofp, p, r);
    fclose(ofp);
  }
  
  fclose(ifp);
  
  /* clean up */
  parti_cleanup();
  MPI_Barrier(groupcomm);
  MPI_Comm_free(&groupcomm);
  MPI_Finalize();

  return 0;
}
