

#include <hdf5.h>
#include <stdlib.h>

#define MAXRANK 6

typedef struct array_t {
  int rank;
  int* size;
} array;







typedef struct region_t {
  int rank;
  int* lo;
  int* up;
  int* st;
} region;


















array* alloc_array(int rank) {
  array* A = 0;
  A = (array*)malloc(sizeof(array));
  A->rank = rank;
  A->size = (int*)calloc(rank, sizeof(int));
  return A;
}


void free_array(array* A) {
  free(A->size);
  free(A);
}


void print_array(array* A) {
  int i;
  printf("rank=%d\n", A->rank);
  for (i = 0; i < A->rank; ++i)
    printf("size[%d]=%d\n", i, A->size[i]);
  printf("\n");
}


region* alloc_region(int rank) {
  region* r = 0;
  r = (region*)malloc(sizeof(region));
  r->rank = rank;
  r->lo = (int*)calloc(rank, sizeof(int));
  r->up = (int*)calloc(rank, sizeof(int));
  r->st = (int*)calloc(rank, sizeof(int));
  return r;
}


void free_region(region* r) {
  free(r->st);
  free(r->up);
  free(r->lo);
  free(r);
}


void print_region(region* r) {
  int i;
  printf("rank=%d\n", r->rank);
  for (i = 0; i < r->rank; ++i) {
    printf("lo[%d]=%d\n", i, r->lo[i]);
    printf("up[%d]=%d\n", i, r->up[i]);
    printf("st[%d]=%d\n", i, r->st[i]);
  }
  printf("\n");
}


char* fill_char(array* A, char value) {
  char* data;
  int i, total = 1;
  
  for (i = 0; i < A->rank; ++i)
    total *= A->size[i];
  data = (char*)calloc(total, sizeof(char));
  for (i = 0; i < total; ++i)
    if (value == 0)
      *(data+i) = (char)0;
    else
      *(data+i) = (char)i;

  return data;
}


int* fill_int(array* A, int value) {
  int* data;
  int i, total = 1;
  
  for (i = 0; i < A->rank; ++i)
    total *= A->size[i];
  data = (int*)calloc(total, sizeof(int));
  for (i = 0; i < total; ++i)
    if (value == 0)
      *(data+i) = (int)0;
    else
      *(data+i) = (int)i;

  return data;
}


float* fill_float(array* A, float value) {
  float* data;
  int i, total = 1;
  
  for (i = 0; i < A->rank; ++i)
    total *= A->size[i];
  data = (float*)calloc(total, sizeof(float));
  for (i = 0; i < total; ++i)
    if (value == 0)
      *(data+i) = (float)0;
    else
      *(data+i) = (float)i;

  return data;
}


double* fill_double(array* A, double value) {
  double* data;
  int i, total = 1;
  
  for (i = 0; i < A->rank; ++i)
    total *= A->size[i];
  data = (double*)calloc(total, sizeof(double));
  for (i = 0; i < total; ++i)
    if (value == 0)
      *(data+i) = (double)0;
    else
      *(data+i) = (double)i;

  return data;
}


void reshape(array* sa, array** nsa, region* sr, region** nsr,
	     array* ra, array** nra, region* rr, region** nrr,
	     int* m, int** nm, int** im) {
  int k, l, n, u = 0;
  
  if (sr->rank > rr->rank) {
    *nsa = alloc_array(sr->rank);
    *nsr = alloc_region(sr->rank);
    *nra = alloc_array(sr->rank);
    *nrr = alloc_region(sr->rank);
    *nm = (int*)calloc(sr->rank, sizeof(int));
    l = rr->rank;
    for (k = 0; k < rr->rank; ++k) {
      (*nsa)->size[k] = sa->size[k];
      (*nsr)->lo[k] = sr->lo[k];
      (*nsr)->up[k] = sr->up[k];
      (*nsr)->st[k] = sr->st[k];
      (*nra)->size[k] = ra->size[k];
      (*nrr)->lo[k] = rr->lo[k];
      (*nrr)->up[k] = rr->up[k];
      (*nrr)->st[k] = rr->st[k];
      if (m[k] != -1)
	(*nm)[k] = m[k];
      else 
	(*nm)[k] = l++;
    }
    for (k = rr->rank; k < sr->rank; ++k) {
      (*nsa)->size[k] = sa->size[k];
      (*nsr)->lo[k] = sr->lo[k];
      (*nsr)->up[k] = sr->up[k];
      (*nsr)->st[k] = sr->st[k];
      (*nra)->size[k] = 1;
      (*nrr)->lo[k] = 0;
      (*nrr)->up[k] = 0;
      (*nrr)->st[k] = 1;
      if (m[k] != -1)
	(*nm)[k] = m[k];
      else 
	(*nm)[k] = l++;
    }
  } else if (rr->rank > sr->rank) {
    (*nsa) = alloc_array(rr->rank);
    (*nsr) = alloc_region(rr->rank);
    (*nra) = alloc_array(rr->rank);
    (*nrr) = alloc_region(rr->rank);
    (*nm) = (int*)calloc(rr->rank, sizeof(int));
    l = sr->rank;
    for (k = 0; k < sr->rank; ++k) {
      (*nsa)->size[k] = sa->size[k];
      (*nsr)->lo[k] = sr->lo[k];
      (*nsr)->up[k] = sr->up[k];
      (*nsr)->st[k] = sr->st[k];
      (*nra)->size[k] = ra->size[k];
      (*nrr)->lo[k] = rr->lo[k];
      (*nrr)->up[k] = rr->up[k];
      (*nrr)->st[k] = rr->st[k];
      (*nm)[k] = m[k];
    }
    for (k = sr->rank; k < rr->rank; ++k) {
      (*nsa)->size[k] = 1;
      (*nsr)->lo[k] = 0;
      (*nsr)->up[k] = 0;
      (*nsr)->st[k] = 1;
      (*nra)->size[k] = ra->size[k];
      (*nrr)->lo[k] = rr->lo[k];
      (*nrr)->up[k] = rr->up[k];
      (*nrr)->st[k] = rr->st[k];
    }
    for (n = 0; n < rr->rank; ++n) {
      for (k = 0; k < sr->rank; ++k)
	if ((*nm)[k] == n)
	  u = 1;
      if (u == 0)
	(*nm)[l++] = n;
      else
	u = 0;
    }
  } else {
    (*nsa) = alloc_array(sr->rank);
    (*nsr) = alloc_region(sr->rank);
    (*nra) = alloc_array(sr->rank);
    (*nrr) = alloc_region(sr->rank);
    (*nm) = (int*)calloc(sr->rank, sizeof(int));
    for (k = 0; k < sr->rank; ++k) {
      (*nsa)->size[k] = sa->size[k];
      (*nsr)->lo[k] = sr->lo[k];
      (*nsr)->up[k] = sr->up[k];
      (*nsr)->st[k] = sr->st[k];
      (*nra)->size[k] = ra->size[k];
      (*nrr)->lo[k] = rr->lo[k];
      (*nrr)->up[k] = rr->up[k];
      (*nrr)->st[k] = rr->st[k];
      (*nm)[k] = m[k];
    }
  }  
  *im = (int*)calloc(sr->rank, sizeof(int));
  for (k = 0; k < (*nsr)->rank; ++k)
    (*im)[(*nm)[k]] = k;
}


void copy_char(char* source, array* sa, region* sr, char* result, array* ra, region* rr, int* m) {
  int i[MAXRANK], j[MAXRANK], *nm, *im;
  array *nsa, *nra;
  region *nsr, *nrr;

  reshape(sa, &nsa, sr, &nsr, ra, &nra, rr, &nrr, m, &nm, &im);
  
  for (i[0] = nsr->lo[0], j[0] = nrr->lo[nm[0]]; 
       i[0] <= nsr->up[0] && j[0] <= nrr->up[nm[0]]; 
       i[0] += nsr->st[0], j[0] += nrr->st[nm[0]]) {
    if (nsr->rank > 1 && nrr->rank > 1) {      
      for (i[1] = nsr->lo[1], j[1] = nrr->lo[nm[1]]; 
	   i[1] <= nsr->up[1] && j[1] <= nrr->up[nm[1]]; 
	   i[1] += nsr->st[1], j[1] += nrr->st[nm[1]]) {
	if (nsr->rank > 2 && nrr->rank > 2) {	  
	  for (i[2] = nsr->lo[2], j[2] = nrr->lo[nm[2]]; 
	       i[2] <= nsr->up[2] && j[2] <= nrr->up[nm[2]]; 
	       i[2] += nsr->st[2], j[2] += nrr->st[nm[2]]) {
	    if (nsr->rank > 3 && nrr->rank > 3) {	      
	      for (i[3] = nsr->lo[3], j[3] = nrr->lo[nm[3]]; 
		   i[3] <= nsr->up[3] && j[3] <= nrr->up[nm[3]]; 
		   i[3] += nsr->st[3], j[3] += nrr->st[nm[3]]) {
		if (nsr->rank > 4 && nrr->rank > 4) {
		  for (i[4] = nsr->lo[4], j[4] = nrr->lo[nm[4]]; 
		       i[4] <= nsr->up[4] && j[4] <= nrr->up[nm[4]]; 
		       i[4] += nsr->st[4], j[4] += nrr->st[nm[4]]) {
		    if (nsr->rank > 5 && nrr->rank > 5) {
		      for (i[5] = nsr->lo[5], j[5] = nrr->lo[nm[5]]; 
			   i[5] <= nsr->up[5] && j[5] <= nrr->up[nm[5]]; 
			   i[5] += nsr->st[5], j[5] += nrr->st[nm[5]]) {
			*(result
			  +j[im[5]]
			  +j[im[4]]*nra->size[5]
			  +j[im[3]]*nra->size[5]*nra->size[4]
			  +j[im[2]]*nra->size[5]*nra->size[4]*nra->size[3]
			  +j[im[1]]*nra->size[5]*nra->size[4]*nra->size[3]*nra->size[2]
			  +j[im[0]]*nra->size[5]*nra->size[4]*nra->size[3]*nra->size[2]*nra->size[1]) 
			  = *(source
			      +i[5]
			      +i[4]*nsa->size[5]
			      +i[3]*nsa->size[5]*nsa->size[4]
			      +i[2]*nsa->size[5]*nsa->size[4]*nsa->size[3]
			      +i[1]*nsa->size[5]*nsa->size[4]*nsa->size[3]*nsa->size[2]
			      +i[0]*nsa->size[5]*nsa->size[4]*nsa->size[3]*nsa->size[2]*nsa->size[1]);
		      }
		    } else {
		      *(result
			+j[im[4]]
			+j[im[3]]*nra->size[4]
			+j[im[2]]*nra->size[4]*nra->size[3]
			+j[im[1]]*nra->size[4]*nra->size[3]*nra->size[2]
			+j[im[0]]*nra->size[4]*nra->size[3]*nra->size[2]*nra->size[1]) 
			= *(source
			    +i[4]
			    +i[3]*nsa->size[4]
			    +i[2]*nsa->size[4]*nsa->size[3]
			    +i[1]*nsa->size[4]*nsa->size[3]*nsa->size[2]
			    +i[0]*nsa->size[4]*nsa->size[3]*nsa->size[2]*nsa->size[1]);
		    }
		  }
		} else {
		  *(result
		    +j[im[3]]
		    +j[im[2]]*nra->size[3]
		    +j[im[1]]*nra->size[3]*nra->size[2]
		    +j[im[0]]*nra->size[3]*nra->size[2]*nra->size[1]) 
		    = *(source
			+i[3]
			+i[2]*nsa->size[3]
			+i[1]*nsa->size[3]*nsa->size[2]
			+i[0]*nsa->size[3]*nsa->size[2]*nsa->size[1]);
		}
	      }
	    } else {
	      *(result
		+j[im[2]]
		+j[im[1]]*nra->size[2]
		+j[im[0]]*nra->size[2]*nra->size[1]) 
		= *(source
		    +i[2]
		    +i[1]*nsa->size[2]
		    +i[0]*nsa->size[2]*nsa->size[1]);
	    }
	  } 
	} else {
	  *(result
	    +j[im[1]]
	    +j[im[0]]*nra->size[1]) 
	    = *(source
		+i[1]
		+i[0]*nsa->size[1]);
	} 
      }
    } else {
      *(result+j[im[0]]) 
	= *(source+i[0]);
    }
  }

  free(im);
  free(nm);
  free_region(nrr);
  free_array(nra);
  free_region(nsr);
  free_array(nsa);
}


void copy_int(int* source, array* sa, region* sr, int* result, array* ra, region* rr, int* m) {
  int i[MAXRANK], j[MAXRANK], *nm, *im;
  array *nsa, *nra;
  region *nsr, *nrr;

  reshape(sa, &nsa, sr, &nsr, ra, &nra, rr, &nrr, m, &nm, &im);
  
  for (i[0] = nsr->lo[0], j[0] = nrr->lo[nm[0]]; 
       i[0] <= nsr->up[0] && j[0] <= nrr->up[nm[0]]; 
       i[0] += nsr->st[0], j[0] += nrr->st[nm[0]]) {
    if (nsr->rank > 1 && nrr->rank > 1) {      
      for (i[1] = nsr->lo[1], j[1] = nrr->lo[nm[1]]; 
	   i[1] <= nsr->up[1] && j[1] <= nrr->up[nm[1]]; 
	   i[1] += nsr->st[1], j[1] += nrr->st[nm[1]]) {
	if (nsr->rank > 2 && nrr->rank > 2) {	  
	  for (i[2] = nsr->lo[2], j[2] = nrr->lo[nm[2]]; 
	       i[2] <= nsr->up[2] && j[2] <= nrr->up[nm[2]]; 
	       i[2] += nsr->st[2], j[2] += nrr->st[nm[2]]) {
	    if (nsr->rank > 3 && nrr->rank > 3) {	      
	      for (i[3] = nsr->lo[3], j[3] = nrr->lo[nm[3]]; 
		   i[3] <= nsr->up[3] && j[3] <= nrr->up[nm[3]]; 
		   i[3] += nsr->st[3], j[3] += nrr->st[nm[3]]) {
		if (nsr->rank > 4 && nrr->rank > 4) {
		  for (i[4] = nsr->lo[4], j[4] = nrr->lo[nm[4]]; 
		       i[4] <= nsr->up[4] && j[4] <= nrr->up[nm[4]]; 
		       i[4] += nsr->st[4], j[4] += nrr->st[nm[4]]) {
		    if (nsr->rank > 5 && nrr->rank > 5) {
		      for (i[5] = nsr->lo[5], j[5] = nrr->lo[nm[5]]; 
			   i[5] <= nsr->up[5] && j[5] <= nrr->up[nm[5]]; 
			   i[5] += nsr->st[5], j[5] += nrr->st[nm[5]]) {
			*(result
			  +j[im[5]]
			  +j[im[4]]*nra->size[5]
			  +j[im[3]]*nra->size[5]*nra->size[4]
			  +j[im[2]]*nra->size[5]*nra->size[4]*nra->size[3]
			  +j[im[1]]*nra->size[5]*nra->size[4]*nra->size[3]*nra->size[2]
			  +j[im[0]]*nra->size[5]*nra->size[4]*nra->size[3]*nra->size[2]*nra->size[1]) 
			  = *(source
			      +i[5]
			      +i[4]*nsa->size[5]
			      +i[3]*nsa->size[5]*nsa->size[4]
			      +i[2]*nsa->size[5]*nsa->size[4]*nsa->size[3]
			      +i[1]*nsa->size[5]*nsa->size[4]*nsa->size[3]*nsa->size[2]
			      +i[0]*nsa->size[5]*nsa->size[4]*nsa->size[3]*nsa->size[2]*nsa->size[1]);
		      }
		    } else {
		      *(result
			+j[im[4]]
			+j[im[3]]*nra->size[4]
			+j[im[2]]*nra->size[4]*nra->size[3]
			+j[im[1]]*nra->size[4]*nra->size[3]*nra->size[2]
			+j[im[0]]*nra->size[4]*nra->size[3]*nra->size[2]*nra->size[1]) 
			= *(source
			    +i[4]
			    +i[3]*nsa->size[4]
			    +i[2]*nsa->size[4]*nsa->size[3]
			    +i[1]*nsa->size[4]*nsa->size[3]*nsa->size[2]
			    +i[0]*nsa->size[4]*nsa->size[3]*nsa->size[2]*nsa->size[1]);
		    }
		  }
		} else {
		  *(result
		    +j[im[3]]
		    +j[im[2]]*nra->size[3]
		    +j[im[1]]*nra->size[3]*nra->size[2]
		    +j[im[0]]*nra->size[3]*nra->size[2]*nra->size[1]) 
		    = *(source
			+i[3]
			+i[2]*nsa->size[3]
			+i[1]*nsa->size[3]*nsa->size[2]
			+i[0]*nsa->size[3]*nsa->size[2]*nsa->size[1]);
		}
	      }
	    } else {
	      *(result
		+j[im[2]]
		+j[im[1]]*nra->size[2]
		+j[im[0]]*nra->size[2]*nra->size[1]) 
		= *(source
		    +i[2]
		    +i[1]*nsa->size[2]
		    +i[0]*nsa->size[2]*nsa->size[1]);
	    }
	  } 
	} else {
	  *(result
	    +j[im[1]]
	    +j[im[0]]*nra->size[1]) 
	    = *(source
		+i[1]
		+i[0]*nsa->size[1]);
	} 
      }
    } else {
      *(result+j[im[0]]) 
	= *(source+i[0]);
    }
  }

  free(im);
  free(nm);
  free_region(nrr);
  free_array(nra);
  free_region(nsr);
  free_array(nsa);
}


void copy_float(float* source, array* sa, region* sr, float* result, array* ra, region* rr, int* m) {
  int i[MAXRANK], j[MAXRANK], *nm, *im;
  array *nsa, *nra;
  region *nsr, *nrr;

  reshape(sa, &nsa, sr, &nsr, ra, &nra, rr, &nrr, m, &nm, &im);
  
  for (i[0] = nsr->lo[0], j[0] = nrr->lo[nm[0]]; 
       i[0] <= nsr->up[0] && j[0] <= nrr->up[nm[0]]; 
       i[0] += nsr->st[0], j[0] += nrr->st[nm[0]]) {
    if (nsr->rank > 1 && nrr->rank > 1) {      
      for (i[1] = nsr->lo[1], j[1] = nrr->lo[nm[1]]; 
	   i[1] <= nsr->up[1] && j[1] <= nrr->up[nm[1]]; 
	   i[1] += nsr->st[1], j[1] += nrr->st[nm[1]]) {
	if (nsr->rank > 2 && nrr->rank > 2) {	  
	  for (i[2] = nsr->lo[2], j[2] = nrr->lo[nm[2]]; 
	       i[2] <= nsr->up[2] && j[2] <= nrr->up[nm[2]]; 
	       i[2] += nsr->st[2], j[2] += nrr->st[nm[2]]) {
	    if (nsr->rank > 3 && nrr->rank > 3) {	      
	      for (i[3] = nsr->lo[3], j[3] = nrr->lo[nm[3]]; 
		   i[3] <= nsr->up[3] && j[3] <= nrr->up[nm[3]]; 
		   i[3] += nsr->st[3], j[3] += nrr->st[nm[3]]) {
		if (nsr->rank > 4 && nrr->rank > 4) {
		  for (i[4] = nsr->lo[4], j[4] = nrr->lo[nm[4]]; 
		       i[4] <= nsr->up[4] && j[4] <= nrr->up[nm[4]]; 
		       i[4] += nsr->st[4], j[4] += nrr->st[nm[4]]) {
		    if (nsr->rank > 5 && nrr->rank > 5) {
		      for (i[5] = nsr->lo[5], j[5] = nrr->lo[nm[5]]; 
			   i[5] <= nsr->up[5] && j[5] <= nrr->up[nm[5]]; 
			   i[5] += nsr->st[5], j[5] += nrr->st[nm[5]]) {
			*(result
			  +j[im[5]]
			  +j[im[4]]*nra->size[5]
			  +j[im[3]]*nra->size[5]*nra->size[4]
			  +j[im[2]]*nra->size[5]*nra->size[4]*nra->size[3]
			  +j[im[1]]*nra->size[5]*nra->size[4]*nra->size[3]*nra->size[2]
			  +j[im[0]]*nra->size[5]*nra->size[4]*nra->size[3]*nra->size[2]*nra->size[1]) 
			  = *(source
			      +i[5]
			      +i[4]*nsa->size[5]
			      +i[3]*nsa->size[5]*nsa->size[4]
			      +i[2]*nsa->size[5]*nsa->size[4]*nsa->size[3]
			      +i[1]*nsa->size[5]*nsa->size[4]*nsa->size[3]*nsa->size[2]
			      +i[0]*nsa->size[5]*nsa->size[4]*nsa->size[3]*nsa->size[2]*nsa->size[1]);
		      }
		    } else {
		      *(result
			+j[im[4]]
			+j[im[3]]*nra->size[4]
			+j[im[2]]*nra->size[4]*nra->size[3]
			+j[im[1]]*nra->size[4]*nra->size[3]*nra->size[2]
			+j[im[0]]*nra->size[4]*nra->size[3]*nra->size[2]*nra->size[1]) 
			= *(source
			    +i[4]
			    +i[3]*nsa->size[4]
			    +i[2]*nsa->size[4]*nsa->size[3]
			    +i[1]*nsa->size[4]*nsa->size[3]*nsa->size[2]
			    +i[0]*nsa->size[4]*nsa->size[3]*nsa->size[2]*nsa->size[1]);
		    }
		  }
		} else {
		  *(result
		    +j[im[3]]
		    +j[im[2]]*nra->size[3]
		    +j[im[1]]*nra->size[3]*nra->size[2]
		    +j[im[0]]*nra->size[3]*nra->size[2]*nra->size[1]) 
		    = *(source
			+i[3]
			+i[2]*nsa->size[3]
			+i[1]*nsa->size[3]*nsa->size[2]
			+i[0]*nsa->size[3]*nsa->size[2]*nsa->size[1]);
		}
	      }
	    } else {
	      *(result
		+j[im[2]]
		+j[im[1]]*nra->size[2]
		+j[im[0]]*nra->size[2]*nra->size[1]) 
		= *(source
		    +i[2]
		    +i[1]*nsa->size[2]
		    +i[0]*nsa->size[2]*nsa->size[1]);
	    }
	  } 
	} else {
	  *(result
	    +j[im[1]]
	    +j[im[0]]*nra->size[1]) 
	    = *(source
		+i[1]
		+i[0]*nsa->size[1]);
	} 
      }
    } else {
      *(result+j[im[0]]) 
	= *(source+i[0]);
    }
  }

  free(im);
  free(nm);
  free_region(nrr);
  free_array(nra);
  free_region(nsr);
  free_array(nsa);
}


void copy_double(double* source, array* sa, region* sr, double* result, array* ra, region* rr, int* m) {
  int i[MAXRANK], j[MAXRANK], *nm, *im;
  array *nsa, *nra;
  region *nsr, *nrr;

  reshape(sa, &nsa, sr, &nsr, ra, &nra, rr, &nrr, m, &nm, &im);
  
  for (i[0] = nsr->lo[0], j[0] = nrr->lo[nm[0]]; 
       i[0] <= nsr->up[0] && j[0] <= nrr->up[nm[0]]; 
       i[0] += nsr->st[0], j[0] += nrr->st[nm[0]]) {
    if (nsr->rank > 1 && nrr->rank > 1) {      
      for (i[1] = nsr->lo[1], j[1] = nrr->lo[nm[1]]; 
	   i[1] <= nsr->up[1] && j[1] <= nrr->up[nm[1]]; 
	   i[1] += nsr->st[1], j[1] += nrr->st[nm[1]]) {
	if (nsr->rank > 2 && nrr->rank > 2) {	  
	  for (i[2] = nsr->lo[2], j[2] = nrr->lo[nm[2]]; 
	       i[2] <= nsr->up[2] && j[2] <= nrr->up[nm[2]]; 
	       i[2] += nsr->st[2], j[2] += nrr->st[nm[2]]) {
	    if (nsr->rank > 3 && nrr->rank > 3) {	      
	      for (i[3] = nsr->lo[3], j[3] = nrr->lo[nm[3]]; 
		   i[3] <= nsr->up[3] && j[3] <= nrr->up[nm[3]]; 
		   i[3] += nsr->st[3], j[3] += nrr->st[nm[3]]) {
		if (nsr->rank > 4 && nrr->rank > 4) {
		  for (i[4] = nsr->lo[4], j[4] = nrr->lo[nm[4]]; 
		       i[4] <= nsr->up[4] && j[4] <= nrr->up[nm[4]]; 
		       i[4] += nsr->st[4], j[4] += nrr->st[nm[4]]) {
		    if (nsr->rank > 5 && nrr->rank > 5) {
		      for (i[5] = nsr->lo[5], j[5] = nrr->lo[nm[5]]; 
			   i[5] <= nsr->up[5] && j[5] <= nrr->up[nm[5]]; 
			   i[5] += nsr->st[5], j[5] += nrr->st[nm[5]]) {
			*(result
			  +j[im[5]]
			  +j[im[4]]*nra->size[5]
			  +j[im[3]]*nra->size[5]*nra->size[4]
			  +j[im[2]]*nra->size[5]*nra->size[4]*nra->size[3]
			  +j[im[1]]*nra->size[5]*nra->size[4]*nra->size[3]*nra->size[2]
			  +j[im[0]]*nra->size[5]*nra->size[4]*nra->size[3]*nra->size[2]*nra->size[1]) 
			  = *(source
			      +i[5]
			      +i[4]*nsa->size[5]
			      +i[3]*nsa->size[5]*nsa->size[4]
			      +i[2]*nsa->size[5]*nsa->size[4]*nsa->size[3]
			      +i[1]*nsa->size[5]*nsa->size[4]*nsa->size[3]*nsa->size[2]
			      +i[0]*nsa->size[5]*nsa->size[4]*nsa->size[3]*nsa->size[2]*nsa->size[1]);
		      }
		    } else {
		      *(result
			+j[im[4]]
			+j[im[3]]*nra->size[4]
			+j[im[2]]*nra->size[4]*nra->size[3]
			+j[im[1]]*nra->size[4]*nra->size[3]*nra->size[2]
			+j[im[0]]*nra->size[4]*nra->size[3]*nra->size[2]*nra->size[1]) 
			= *(source
			    +i[4]
			    +i[3]*nsa->size[4]
			    +i[2]*nsa->size[4]*nsa->size[3]
			    +i[1]*nsa->size[4]*nsa->size[3]*nsa->size[2]
			    +i[0]*nsa->size[4]*nsa->size[3]*nsa->size[2]*nsa->size[1]);
		    }
		  }
		} else {
		  *(result
		    +j[im[3]]
		    +j[im[2]]*nra->size[3]
		    +j[im[1]]*nra->size[3]*nra->size[2]
		    +j[im[0]]*nra->size[3]*nra->size[2]*nra->size[1]) 
		    = *(source
			+i[3]
			+i[2]*nsa->size[3]
			+i[1]*nsa->size[3]*nsa->size[2]
			+i[0]*nsa->size[3]*nsa->size[2]*nsa->size[1]);
		}
	      }
	    } else {
	      *(result
		+j[im[2]]
		+j[im[1]]*nra->size[2]
		+j[im[0]]*nra->size[2]*nra->size[1]) 
		= *(source
		    +i[2]
		    +i[1]*nsa->size[2]
		    +i[0]*nsa->size[2]*nsa->size[1]);
	    }
	  } 
	} else {
	  *(result
	    +j[im[1]]
	    +j[im[0]]*nra->size[1]) 
	    = *(source
		+i[1]
		+i[0]*nsa->size[1]);
	} 
      }
    } else {
      *(result+j[im[0]]) 
	= *(source+i[0]);
    }
  }

  free(im);
  free(nm);
  free_region(nrr);
  free_array(nra);
  free_region(nsr);
  free_array(nsa);
}


void create_char(char* filename, array* A, array* B, region* rA, region* rB, int* map) {
  hid_t file, dataset, dataspace, datatype, group, attribute;
  hsize_t rank, size[MAXRANK];
  herr_t sts;
  char* source, *result;
  int i;

  file = H5Fcreate(filename, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
  datatype = H5Tcopy(H5T_NATIVE_CHAR);
  sts = H5Tset_order(datatype, H5T_ORDER_LE);

  for (i = 0; i < A->rank; ++i)
    size[i] = (hsize_t)A->size[i];
  dataspace = H5Screate_simple(A->rank, size, NULL);
  dataset = H5Dcreate(file, "/source", datatype, dataspace, H5P_DEFAULT);
  source = fill_char(A, 1);
  sts = H5Dwrite(dataset, H5T_NATIVE_CHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, source);
  H5Dclose(dataset);
  H5Sclose(dataspace);

  for (i = 0; i < B->rank; ++i)
    size[i] = (hsize_t)B->size[i];
  dataspace = H5Screate_simple(B->rank, size, NULL);
  dataset = H5Dcreate(file, "/result", datatype, dataspace, H5P_DEFAULT);
  result = fill_char(B, 0);
  copy_char(source, A, rA, result, B, rB, map);
  sts = H5Dwrite(dataset, H5T_NATIVE_CHAR, H5S_ALL, H5S_ALL, H5P_DEFAULT, result);
  H5Dclose(dataset);  
  H5Sclose(dataspace);

  free(source);
  free(result);

  H5Tclose(datatype);

  datatype = H5Tcopy(H5T_NATIVE_INT);
  sts = H5Tset_order(datatype, H5T_ORDER_LE);
  group = H5Gopen(file, "/");

  rank = (hsize_t)rA->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "sourceLower", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rA->lo);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rA->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "sourceUpper", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rA->up);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rA->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "sourceStride", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rA->st);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rB->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "resultLower", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rB->lo);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rB->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "resultUpper", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rB->up);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rB->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "resultStride", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rB->st);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rA->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "mapping", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, map);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  H5Gclose(group);
  H5Tclose(datatype);
  
  H5Fclose(file);
}


void create_int(char* filename, array* A, array* B, region* rA, region* rB, int* map) {
  hid_t file, dataset, dataspace, datatype, group, attribute;
  hsize_t rank, size[MAXRANK];
  herr_t sts;
  int* source, *result;
  int i;

  file = H5Fcreate(filename, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
  datatype = H5Tcopy(H5T_NATIVE_INT);
  sts = H5Tset_order(datatype, H5T_ORDER_LE);

  for (i = 0; i < A->rank; ++i)
    size[i] = (hsize_t)A->size[i];
  dataspace = H5Screate_simple(A->rank, size, NULL);
  dataset = H5Dcreate(file, "/source", datatype, dataspace, H5P_DEFAULT);
  source = fill_int(A, 1);
  sts = H5Dwrite(dataset, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, source);
  H5Dclose(dataset);
  H5Sclose(dataspace);

  for (i = 0; i < B->rank; ++i)
    size[i] = (hsize_t)B->size[i];
  dataspace = H5Screate_simple(B->rank, size, NULL);
  dataset = H5Dcreate(file, "/result", datatype, dataspace, H5P_DEFAULT);
  result = fill_int(B, 0);
  copy_int(source, A, rA, result, B, rB, map);
  sts = H5Dwrite(dataset, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, result);
  H5Dclose(dataset);  
  H5Sclose(dataspace);

  free(source);
  free(result);

  H5Tclose(datatype);

  datatype = H5Tcopy(H5T_NATIVE_INT);
  sts = H5Tset_order(datatype, H5T_ORDER_LE);
  group = H5Gopen(file, "/");

  rank = (hsize_t)rA->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "sourceLower", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rA->lo);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rA->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "sourceUpper", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rA->up);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rA->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "sourceStride", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rA->st);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rB->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "resultLower", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rB->lo);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rB->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "resultUpper", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rB->up);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rB->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "resultStride", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rB->st);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rA->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "mapping", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, map);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  H5Gclose(group);
  H5Tclose(datatype);
  
  H5Fclose(file);
}


void create_float(char* filename, array* A, array* B, region* rA, region* rB, int* map) {
  hid_t file, dataset, dataspace, datatype, group, attribute;
  hsize_t rank, size[MAXRANK];
  herr_t sts;
  float* source, *result;
  int i;

  file = H5Fcreate(filename, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
  datatype = H5Tcopy(H5T_NATIVE_FLOAT);
  sts = H5Tset_order(datatype, H5T_ORDER_LE);

  for (i = 0; i < A->rank; ++i)
    size[i] = (hsize_t)A->size[i];
  dataspace = H5Screate_simple(A->rank, size, NULL);
  dataset = H5Dcreate(file, "/source", datatype, dataspace, H5P_DEFAULT);
  source = fill_float(A, 1);
  sts = H5Dwrite(dataset, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, source);
  H5Dclose(dataset);
  H5Sclose(dataspace);

  for (i = 0; i < B->rank; ++i)
    size[i] = (hsize_t)B->size[i];
  dataspace = H5Screate_simple(B->rank, size, NULL);
  dataset = H5Dcreate(file, "/result", datatype, dataspace, H5P_DEFAULT);
  result = fill_float(B, 0);
  copy_float(source, A, rA, result, B, rB, map);
  sts = H5Dwrite(dataset, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, result);
  H5Dclose(dataset);  
  H5Sclose(dataspace);

  free(source);
  free(result);

  H5Tclose(datatype);

  datatype = H5Tcopy(H5T_NATIVE_INT);
  sts = H5Tset_order(datatype, H5T_ORDER_LE);
  group = H5Gopen(file, "/");

  rank = (hsize_t)rA->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "sourceLower", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rA->lo);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rA->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "sourceUpper", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rA->up);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rA->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "sourceStride", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rA->st);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rB->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "resultLower", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rB->lo);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rB->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "resultUpper", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rB->up);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rB->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "resultStride", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rB->st);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rA->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "mapping", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, map);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  H5Gclose(group);
  H5Tclose(datatype);
  
  H5Fclose(file);
}


void create_double(char* filename, array* A, array* B, region* rA, region* rB, int* map) {
  hid_t file, dataset, dataspace, datatype, group, attribute;
  hsize_t rank, size[MAXRANK];
  herr_t sts;
  double* source, *result;
  int i;

  file = H5Fcreate(filename, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
  datatype = H5Tcopy(H5T_NATIVE_DOUBLE);
  sts = H5Tset_order(datatype, H5T_ORDER_LE);

  for (i = 0; i < A->rank; ++i)
    size[i] = (hsize_t)A->size[i];
  dataspace = H5Screate_simple(A->rank, size, NULL);
  dataset = H5Dcreate(file, "/source", datatype, dataspace, H5P_DEFAULT);
  source = fill_double(A, 1);
  sts = H5Dwrite(dataset, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, source);
  H5Dclose(dataset);
  H5Sclose(dataspace);

  for (i = 0; i < B->rank; ++i)
    size[i] = (hsize_t)B->size[i];
  dataspace = H5Screate_simple(B->rank, size, NULL);
  dataset = H5Dcreate(file, "/result", datatype, dataspace, H5P_DEFAULT);
  result = fill_double(B, 0);
  copy_double(source, A, rA, result, B, rB, map);
  sts = H5Dwrite(dataset, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, result);
  H5Dclose(dataset);  
  H5Sclose(dataspace);

  free(source);
  free(result);

  H5Tclose(datatype);

  datatype = H5Tcopy(H5T_NATIVE_INT);
  sts = H5Tset_order(datatype, H5T_ORDER_LE);
  group = H5Gopen(file, "/");

  rank = (hsize_t)rA->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "sourceLower", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rA->lo);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rA->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "sourceUpper", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rA->up);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rA->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "sourceStride", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rA->st);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rB->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "resultLower", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rB->lo);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rB->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "resultUpper", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rB->up);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rB->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "resultStride", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, rB->st);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  rank = (hsize_t)rA->rank;
  dataspace = H5Screate_simple(1, &rank, NULL);
  attribute = H5Acreate(group, "mapping", datatype, dataspace, H5P_DEFAULT);
  sts = H5Awrite(attribute, H5T_NATIVE_INT, map);
  sts = H5Aclose(attribute);
  H5Sclose(dataspace);

  H5Gclose(group);
  H5Tclose(datatype);
  
  H5Fclose(file);
}


int main(int argc, char* argv[]) {
  int i, r1, r2, c = 1;
  char* p, ta1[40], ta2[40], tb1[40], tb2[40], m[40];
  array* A, *B;
  region* rA, *rB;
  int *map;
  char filename[80];

  while (fscanf(stdin, "%d %s %s -> %d %s %s %s", &r1, ta1, ta2, &r2, tb1, tb2, m) == 7) {
    printf("R1:%d TA1:%s TA2:%s R2:%d TB1:%s TB2:%s M:%s\n", r1, ta1, ta2, r2, tb1, tb2, m);
    A = alloc_array(r1);
    p = &ta1[0];
    for (i = 0; i < r1; ++i) {
      A->size[i] = atoi(p);
      p = strchr(p, 'x'); p++;
    }

    rA = alloc_region(r1);
    p = &ta2[0];
    for (i = 0; i < r1; ++i) {
      rA->lo[i] = atoi(p);
      p = strchr(p, ','); p++;
      rA->up[i] = atoi(p);
      p = strchr(p, ','); p++;
      rA->st[i] = atoi(p);
      p = strchr(p, ';'); p++;
    }
    
    B = alloc_array(r2);
    p = &tb1[0];
    for (i = 0; i < r2; ++i) {
      B->size[i] = atoi(p);
      p = strchr(p, 'x'); p++;
    }      
    
    rB = alloc_region(r2);
    p = &tb2[0];
    for (i = 0; i < r2; ++i) {
      rB->lo[i] = atoi(p);
      p = strchr(p, ','); p++;
      rB->up[i] = atoi(p);
      p = strchr(p, ','); p++;
      rB->st[i] = atoi(p);
      p = strchr(p, ';'); p++; 
    }

    map = (int*)calloc(r1, sizeof(int));
    p = &m[0];
    for (i = 0; i < r1; ++i) {
      map[i] = atoi(p);
      p = strchr(p, ','); p++;
    }

    sprintf(filename, "char%dD%dc%d.h5", r1, r2, c);
    create_char(filename, A, B, rA, rB, map);
    sprintf(filename, "int%dD%dc%d.h5", r1, r2, c);
    create_int(filename, A, B, rA, rB, map);
    sprintf(filename, "float%dD%dc%d.h5", r1, r2, c);
    create_float(filename, A, B, rA, rB, map);
    sprintf(filename, "double%dD%dc%d.h5", r1, r2, c);
    create_double(filename, A, B, rA, rB, map);
    
    free_region(rB);
    free_array(B);
    free_region(rA);
    free_array(A);
    
    ++c;
  }
}

