

#include <multiblockparti.h>
#include <schedules.h>
#include <partitioner.h>
#include <reader.h>
#include <hdf5.h>
#include <stdlib.h>
#include "utilities.h"








int source_params(char* arrayfile, parti_region_t* region1, parti_region_t* region2, int map[MAX_DIM]) {
  hid_t file, group, attribute;
  herr_t sts;

  /* open the file */
  file = H5Fopen(arrayfile, H5F_ACC_RDONLY, H5P_DEFAULT);
  if (file < 0) {
    fprintf(stderr, "hdf5 file not found\n");
    return 0;
  }

  /* get the group */
  group = H5Gopen(file, "/");
  if (group < 0) {
    fprintf(stderr, "group not found\n");
    return 0;
  }
  
  /* get the attributes */
  attribute = H5Aopen_name(group, "sourceLower");
  if (attribute < 0) {
    fprintf(stderr, "attribute not found\n");
    return 0;
  }
  sts = H5Aread(attribute, H5T_NATIVE_INT, region1->lower);
  sts = H5Aclose(attribute);
  attribute = H5Aopen_name(group, "sourceUpper");
  if (attribute < 0) {
    fprintf(stderr, "attribute not found\n");
    return 0;
  }
  sts = H5Aread(attribute, H5T_NATIVE_INT, region1->upper);
  sts = H5Aclose(attribute);
  attribute = H5Aopen_name(group, "sourceStride");
  if (attribute < 0) {
    fprintf(stderr, "attribute not found\n");
    return 0;
  }
  sts = H5Aread(attribute, H5T_NATIVE_INT, region1->stride);
  sts = H5Aclose(attribute);
  attribute = H5Aopen_name(group, "resultLower");
  if (attribute < 0) {
    fprintf(stderr, "attribute not found\n");
    return 0;
  }
  sts = H5Aread(attribute, H5T_NATIVE_INT, region2->lower);
  sts = H5Aclose(attribute);
  attribute = H5Aopen_name(group, "resultUpper");
  if (attribute < 0) {
    fprintf(stderr, "attribute not found\n");
    return 0;
  }
  sts = H5Aread(attribute, H5T_NATIVE_INT, region2->upper);
  sts = H5Aclose(attribute);
  attribute = H5Aopen_name(group, "resultStride");
  if (attribute < 0) {
    fprintf(stderr, "attribute not found\n");
    return 0;
  }
  sts = H5Aread(attribute, H5T_NATIVE_INT, region2->stride);
  sts = H5Aclose(attribute);
  attribute = H5Aopen_name(group, "mapping");
  if (attribute < 0) {
    fprintf(stderr, "attribute not found\n");
    return 0;
  }
  sts = H5Aread(attribute, H5T_NATIVE_INT, map);
  sts = H5Aclose(attribute);
  
  sts = H5Gclose(group);
  sts = H5Fclose(file);
  return 1;
}


int inter_test2char(char* partition1, char* arrayfile, char* partition2) {
  int i;
  
  parti_decomp_t* decomp1, *decomp2;
  parti_darray_t* darray1, *darray2;
  
  int rank, size[MAX_DIM];
  int igcl[MAX_DIM] = {0, 0, 0, 0, 0, 0};
  int igcu[MAX_DIM] = {0, 0, 0, 0, 0, 0};  
  int cnt;

  char* A, *B; /* local portion of the arrays */

  int map[MAX_DIM];
  parti_region_t* region1, *region2;
  parti_sched_t* sched;


  if (!(decomp1 = parse_document(partition1))) {
    fprintf(stderr, "decomp1 creation failed\n");
    return 0;
  }
  
  if (!(darray1 = parti_ialign(decomp1, igcl, igcu))) {
    fprintf(stderr, "darray1 creation failed\n");
    parti_free_decomp(decomp1);
    return 1;
  }
  
  A = (char*)source_array(arrayfile, "/source", darray1);

  cnt = parti_local_size(darray1, size);
#if PARTI_DEBUG > 0
  fprint_array(stdout, 'c', A, darray1->decomp->rank, size, 0, 0);
#endif
  
  if (!(decomp2 = parse_document(partition2))) {
    fprintf(stderr, "decomp2 creation failed\n");
    parti_free_decomp(decomp1);
    parti_free_darray(darray1);
    free(A);
    return 0;
  }
  
  if (!(darray2 = parti_ialign(decomp2, igcl, igcu))) {
    fprintf(stderr, "darray2 creation failed\n");
    parti_free_decomp(decomp1);
    parti_free_decomp(decomp2);
    parti_free_darray(darray1);
    free(A);
    return 0;
  }
  
  cnt = parti_local_size(darray2, size);
  B = (char*)malloc(cnt*sizeof(char));
  
  elemset('c', B, 0, darray2->decomp->rank, size, 0, 0);
#if PARTI_DEBUG > 0
  fprint_array(stdout, 'c', B, darray2->decomp->rank, size, 0, 0);
#endif
  
  region1 = parti_alloc_region(darray1->decomp->rank);
  region2 = parti_alloc_region(darray2->decomp->rank);
  if (!(source_params(arrayfile, region1, region2, map))) {
    fprintf(stderr, "could not get move parameters\n");
    parti_free_region(region1);
    parti_free_region(region2);    
    parti_free_decomp(decomp1);
    parti_free_decomp(decomp2);
    parti_free_darray(darray1);
    parti_free_darray(darray2);
    free(A);
    free(B);
    return 0;
  }

#if PARTI_DEBUG > 1
  print_region(region1);
  print_region(region2);
#endif
  
  sched = parti_region_sched(darray1, region1, darray2, region2, map); 
  parti_data_move_char(A, sched, B);
  
#if PARTI_DEBUG > 0
  fprint_array(stdout, 'c', B, darray2->decomp->rank, size, 0, 0);
#endif
  
  parti_free_region(region1);
  parti_free_region(region2);
  
  free(A);
  
  A = (char*)source_array(arrayfile, "/result", darray2);
  
#if PARTI_DEBUG > 0
  fprint_array(stdout, 'c', A, darray2->decomp->rank, size, 0, 0);
#endif
  
  for (i = 0; i < cnt; ++i) {
    if (*(A+i) != *(B+i)) {
      fprintf(stderr, "comparison failed\n");
      return 0;
    }
  }

  free(B);  
  parti_free_darray(darray2);
  parti_free_decomp(decomp2);
  
  free(A);
  parti_free_darray(darray1);
  parti_free_decomp(decomp1);

  return 1;
}


int inter_test2int(char* partition1, char* arrayfile, char* partition2) {
  int i;
  
  parti_decomp_t* decomp1, *decomp2;
  parti_darray_t* darray1, *darray2;
  
  int rank, size[MAX_DIM];
  int igcl[MAX_DIM] = {0, 0, 0, 0, 0, 0};
  int igcu[MAX_DIM] = {0, 0, 0, 0, 0, 0};  
  int cnt;

  int* A, *B; /* local portion of the arrays */

  int map[MAX_DIM];
  parti_region_t* region1, *region2;
  parti_sched_t* sched;


  if (!(decomp1 = parse_document(partition1))) {
    fprintf(stderr, "decomp1 creation failed\n");
    return 0;
  }
  
  if (!(darray1 = parti_ialign(decomp1, igcl, igcu))) {
    fprintf(stderr, "darray1 creation failed\n");
    parti_free_decomp(decomp1);
    return 1;
  }
  
  A = (int*)source_array(arrayfile, "/source", darray1);

  cnt = parti_local_size(darray1, size);
#if PARTI_DEBUG > 0
  fprint_array(stdout, 'i', A, darray1->decomp->rank, size, 0, 0);
#endif
  
  if (!(decomp2 = parse_document(partition2))) {
    fprintf(stderr, "decomp2 creation failed\n");
    parti_free_decomp(decomp1);
    parti_free_darray(darray1);
    free(A);
    return 0;
  }
  
  if (!(darray2 = parti_ialign(decomp2, igcl, igcu))) {
    fprintf(stderr, "darray2 creation failed\n");
    parti_free_decomp(decomp1);
    parti_free_decomp(decomp2);
    parti_free_darray(darray1);
    free(A);
    return 0;
  }
  
  cnt = parti_local_size(darray2, size);
  B = (int*)malloc(cnt*sizeof(int));
  
  elemset('i', B, 0, darray2->decomp->rank, size, 0, 0);
#if PARTI_DEBUG > 0
  fprint_array(stdout, 'i', B, darray2->decomp->rank, size, 0, 0);
#endif
  
  region1 = parti_alloc_region(darray1->decomp->rank);
  region2 = parti_alloc_region(darray2->decomp->rank);
  if (!(source_params(arrayfile, region1, region2, map))) {
    fprintf(stderr, "could not get move parameters\n");
    parti_free_region(region1);
    parti_free_region(region2);    
    parti_free_decomp(decomp1);
    parti_free_decomp(decomp2);
    parti_free_darray(darray1);
    parti_free_darray(darray2);
    free(A);
    free(B);
    return 0;
  }

#if PARTI_DEBUG > 1
  print_region(region1);
  print_region(region2);
#endif
  
  sched = parti_region_sched(darray1, region1, darray2, region2, map); 
  parti_data_move_int(A, sched, B);
  
#if PARTI_DEBUG > 0
  fprint_array(stdout, 'i', B, darray2->decomp->rank, size, 0, 0);
#endif
  
  parti_free_region(region1);
  parti_free_region(region2);
  
  free(A);
  
  A = (int*)source_array(arrayfile, "/result", darray2);
  
#if PARTI_DEBUG > 0
  fprint_array(stdout, 'i', A, darray2->decomp->rank, size, 0, 0);
#endif
  
  for (i = 0; i < cnt; ++i) {
    if (*(A+i) != *(B+i)) {
      fprintf(stderr, "comparison failed\n");
      return 0;
    }
  }

  free(B);  
  parti_free_darray(darray2);
  parti_free_decomp(decomp2);
  
  free(A);
  parti_free_darray(darray1);
  parti_free_decomp(decomp1);

  return 1;
}


int inter_test2float(char* partition1, char* arrayfile, char* partition2) {
  int i;
  
  parti_decomp_t* decomp1, *decomp2;
  parti_darray_t* darray1, *darray2;
  
  int rank, size[MAX_DIM];
  int igcl[MAX_DIM] = {0, 0, 0, 0, 0, 0};
  int igcu[MAX_DIM] = {0, 0, 0, 0, 0, 0};  
  int cnt;

  float* A, *B; /* local portion of the arrays */

  int map[MAX_DIM];
  parti_region_t* region1, *region2;
  parti_sched_t* sched;


  if (!(decomp1 = parse_document(partition1))) {
    fprintf(stderr, "decomp1 creation failed\n");
    return 0;
  }
  
  if (!(darray1 = parti_ialign(decomp1, igcl, igcu))) {
    fprintf(stderr, "darray1 creation failed\n");
    parti_free_decomp(decomp1);
    return 1;
  }
  
  A = (float*)source_array(arrayfile, "/source", darray1);

  cnt = parti_local_size(darray1, size);
#if PARTI_DEBUG > 0
  fprint_array(stdout, 'f', A, darray1->decomp->rank, size, 0, 0);
#endif
  
  if (!(decomp2 = parse_document(partition2))) {
    fprintf(stderr, "decomp2 creation failed\n");
    parti_free_decomp(decomp1);
    parti_free_darray(darray1);
    free(A);
    return 0;
  }
  
  if (!(darray2 = parti_ialign(decomp2, igcl, igcu))) {
    fprintf(stderr, "darray2 creation failed\n");
    parti_free_decomp(decomp1);
    parti_free_decomp(decomp2);
    parti_free_darray(darray1);
    free(A);
    return 0;
  }
  
  cnt = parti_local_size(darray2, size);
  B = (float*)malloc(cnt*sizeof(float));
  
  elemset('f', B, 0, darray2->decomp->rank, size, 0, 0);
#if PARTI_DEBUG > 0
  fprint_array(stdout, 'f', B, darray2->decomp->rank, size, 0, 0);
#endif
  
  region1 = parti_alloc_region(darray1->decomp->rank);
  region2 = parti_alloc_region(darray2->decomp->rank);
  if (!(source_params(arrayfile, region1, region2, map))) {
    fprintf(stderr, "could not get move parameters\n");
    parti_free_region(region1);
    parti_free_region(region2);    
    parti_free_decomp(decomp1);
    parti_free_decomp(decomp2);
    parti_free_darray(darray1);
    parti_free_darray(darray2);
    free(A);
    free(B);
    return 0;
  }

#if PARTI_DEBUG > 1
  print_region(region1);
  print_region(region2);
#endif
  
  sched = parti_region_sched(darray1, region1, darray2, region2, map); 
  parti_data_move_float(A, sched, B);
  
#if PARTI_DEBUG > 0
  fprint_array(stdout, 'f', B, darray2->decomp->rank, size, 0, 0);
#endif
  
  parti_free_region(region1);
  parti_free_region(region2);
  
  free(A);
  
  A = (float*)source_array(arrayfile, "/result", darray2);
  
#if PARTI_DEBUG > 0
  fprint_array(stdout, 'f', A, darray2->decomp->rank, size, 0, 0);
#endif
  
  for (i = 0; i < cnt; ++i) {
    if (*(A+i) != *(B+i)) {
      fprintf(stderr, "comparison failed\n");
      return 0;
    }
  }

  free(B);  
  parti_free_darray(darray2);
  parti_free_decomp(decomp2);
  
  free(A);
  parti_free_darray(darray1);
  parti_free_decomp(decomp1);

  return 1;
}


int inter_test2double(char* partition1, char* arrayfile, char* partition2) {
  int i;
  
  parti_decomp_t* decomp1, *decomp2;
  parti_darray_t* darray1, *darray2;
  
  int rank, size[MAX_DIM];
  int igcl[MAX_DIM] = {0, 0, 0, 0, 0, 0};
  int igcu[MAX_DIM] = {0, 0, 0, 0, 0, 0};  
  int cnt;

  double* A, *B; /* local portion of the arrays */

  int map[MAX_DIM];
  parti_region_t* region1, *region2;
  parti_sched_t* sched;


  if (!(decomp1 = parse_document(partition1))) {
    fprintf(stderr, "decomp1 creation failed\n");
    return 0;
  }
  
  if (!(darray1 = parti_ialign(decomp1, igcl, igcu))) {
    fprintf(stderr, "darray1 creation failed\n");
    parti_free_decomp(decomp1);
    return 1;
  }
  
  A = (double*)source_array(arrayfile, "/source", darray1);

  cnt = parti_local_size(darray1, size);
#if PARTI_DEBUG > 0
  fprint_array(stdout, 'd', A, darray1->decomp->rank, size, 0, 0);
#endif
  
  if (!(decomp2 = parse_document(partition2))) {
    fprintf(stderr, "decomp2 creation failed\n");
    parti_free_decomp(decomp1);
    parti_free_darray(darray1);
    free(A);
    return 0;
  }
  
  if (!(darray2 = parti_ialign(decomp2, igcl, igcu))) {
    fprintf(stderr, "darray2 creation failed\n");
    parti_free_decomp(decomp1);
    parti_free_decomp(decomp2);
    parti_free_darray(darray1);
    free(A);
    return 0;
  }
  
  cnt = parti_local_size(darray2, size);
  B = (double*)malloc(cnt*sizeof(double));
  
  elemset('d', B, 0, darray2->decomp->rank, size, 0, 0);
#if PARTI_DEBUG > 0
  fprint_array(stdout, 'd', B, darray2->decomp->rank, size, 0, 0);
#endif
  
  region1 = parti_alloc_region(darray1->decomp->rank);
  region2 = parti_alloc_region(darray2->decomp->rank);
  if (!(source_params(arrayfile, region1, region2, map))) {
    fprintf(stderr, "could not get move parameters\n");
    parti_free_region(region1);
    parti_free_region(region2);    
    parti_free_decomp(decomp1);
    parti_free_decomp(decomp2);
    parti_free_darray(darray1);
    parti_free_darray(darray2);
    free(A);
    free(B);
    return 0;
  }

#if PARTI_DEBUG > 1
  print_region(region1);
  print_region(region2);
#endif
  
  sched = parti_region_sched(darray1, region1, darray2, region2, map); 
  parti_data_move_double(A, sched, B);
  
#if PARTI_DEBUG > 0
  fprint_array(stdout, 'd', B, darray2->decomp->rank, size, 0, 0);
#endif
  
  parti_free_region(region1);
  parti_free_region(region2);
  
  free(A);
  
  A = (double*)source_array(arrayfile, "/result", darray2);
  
#if PARTI_DEBUG > 0
  fprint_array(stdout, 'd', A, darray2->decomp->rank, size, 0, 0);
#endif
  
  for (i = 0; i < cnt; ++i) {
    if (*(A+i) != *(B+i)) {
      fprintf(stderr, "comparison failed\n");
      return 0;
    }
  }

  free(B);  
  parti_free_darray(darray2);
  parti_free_decomp(decomp2);
  
  free(A);
  parti_free_darray(darray1);
  parti_free_decomp(decomp1);

  return 1;
}


void parti_test(FILE* ifp, FILE* ofp, int p, int r) {
  int sts, test = 1;
  char xml1[80], hdf[80], xml2[80];
  char* c;
  
  while (fscanf(ifp, "%s %s %s", xml1, hdf, xml2) == 3) {
    fprintf(ofp, "task %d running test %d...", r, test);
    
    if (c = (char*)strrchr(hdf, '/'))
      ++c;
    else 
      c = hdf;
    
    switch (*c) {
    case 'c':
      sts = inter_test2char(xml1, hdf, xml2);
      break;
    case 'i':
      sts = inter_test2int(xml1, hdf, xml2);
      break;
    case 'f':
      sts = inter_test2float(xml1, hdf, xml2);
      break;
    case 'd':
      sts = inter_test2double(xml1, hdf, xml2);
      break;
    default:
      sts = 0;
    }
    
    if (sts == 0) {
      fprintf(ofp, "failed\n");
    } else {
      fprintf(ofp, "succeeded\n");
    }
    
    PARTI_gsync();
    ++test;
  }
}

