#include <multiblockparti.h>
#include <partitioner.h>
#include <birreg.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include "utilities.h"

#define min(A,B) ((A)<(B)?(A):(B))
#define max(A,B) ((A)>(B)?(A):(B))

int check_element(parti_darray_t* darray, parti_region_t* all, parti_region_t* one, int* A, int index, int depth) {
  decomp_block_t** blocks = 0;
  int i, count = 0;
  
  if (darray->decomp->rank == depth) {
#ifdef PARTI_DEBUG
    print_region(one); 
#endif
    get_region_blocks(one, &(darray->decomp->root), &blocks, &count);
#ifdef PARTI_DEBUG
    printf("block count:%d\n", count);
#endif
    assert(count == 1); /* one ghost cell should be owned by one and only one task */
#ifdef PARTI_DEBUG
    printf("actual:%d expected:%d\n", *(A+index), (*blocks)->task);
#endif
    if (*(A+index) != (*blocks)->task) {
      free(blocks);
      return 0;
    } else {
      free(blocks);
      return 1;
    }
  }
  
  for (i = all->lower[depth]; i <= all->upper[depth]; ++i) {
    one->lower[depth] = one->upper[depth] = parti_local_to_global(darray, i, depth, 1);
    if (check_element(darray, all, one, A, (index*all->stride[depth])+i, depth+1) == 0)
      return 0;
  }
  
  return 1;
}

int verify_move(parti_darray_t* darray, int* A, int* fill, int type) {
  int i, j, k, task;
  parti_region_t* allGhost, *oneGhost;
  
  task = PARTI_myproc();
  
  /* get a local bounding box */
  allGhost = parti_alloc_region(darray->decomp->rank);
  switch (type) {
  case 0:
    for (j = 0; j < darray->decomp->rank; ++j) {
      if (fill[j] < 0) {
	allGhost->lower[j] = parti_global_lower_bound(darray, j) + fill[j];
	allGhost->upper[j] = parti_global_lower_bound(darray, j) - 1;
      } else if (fill[j] > 0) {
	allGhost->lower[j] = parti_global_upper_bound(darray, j) + 1;
	allGhost->upper[j] = parti_global_upper_bound(darray, j) + fill[j];
      } else {
	allGhost->lower[j] = parti_global_lower_bound(darray, j);
	allGhost->upper[j] = parti_global_upper_bound(darray, j);
      }
      if (allGhost->lower[j] < 0 || allGhost->upper[j] >= darray->decomp->size[j]) {
	parti_free_region(allGhost);
	return 1;
      }
      allGhost->lower[j] = parti_global_to_local(darray, allGhost->lower[j], j, 1);
      allGhost->upper[j] = parti_global_to_local(darray, allGhost->upper[j], j, 1);
    }
    break;
  case 1:
    for (j = 0; j < darray->decomp->rank; ++j) {
      if (fill[j] < 0) {
	allGhost->lower[j] = parti_global_lower_bound(darray, j) + fill[j];
	allGhost->upper[j] = parti_global_upper_bound(darray, j);
      } else if (fill[j] > 0) {
	allGhost->lower[j] = parti_global_lower_bound(darray, j);
	allGhost->upper[j] = parti_global_upper_bound(darray, j) + fill[j];
      } else {
	allGhost->lower[j] = parti_global_lower_bound(darray, j);
	allGhost->upper[j] = parti_global_upper_bound(darray, j);
      }
      allGhost->lower[j] = max(allGhost->lower[j], 0);
      allGhost->upper[j] = min(allGhost->upper[j], darray->decomp->size[j]-1);
      allGhost->lower[j] = parti_global_to_local(darray, allGhost->lower[j], j, 1);
      allGhost->upper[j] = parti_global_to_local(darray, allGhost->upper[j], j, 1);
    }
    break;
  case 2:
    parti_local_size(darray, allGhost->stride);
    for (j = 0; j < darray->decomp->rank; ++j) {
      allGhost->lower[j] = max(parti_local_to_global(darray, 0, j, 1), 0);
      allGhost->upper[j] = min(parti_local_to_global(darray, allGhost->stride[j]-1, j, 1), darray->decomp->size[j]-1);
      allGhost->lower[j] = parti_global_to_local(darray, allGhost->lower[j], j, 1);
      allGhost->upper[j] = parti_global_to_local(darray, allGhost->upper[j], j, 1);
    }
    break;
  }
  parti_local_size(darray, allGhost->stride);
  
  /* look at each element and insure that it has changed to the value at the source */
  oneGhost = parti_alloc_region(darray->decomp->rank);
  if (check_element(darray, allGhost, oneGhost, A, 0, 0) == 0) {
    parti_free_region(oneGhost);
    parti_free_region(allGhost);
    return 0;
  }

  parti_free_region(oneGhost);
  parti_free_region(allGhost);
  return 1;
}

int intra_test(FILE* ofp, char* xml, int* igcl, int* igcu, int* fill) {
  parti_decomp_t* decomp;
  parti_darray_t* darray;
  int i, sts;
  
  int size[3];
  int blk[3], cnt;

  int* A; /* local portion of the array */
  
  parti_sched_t* sched;

  static int count = 0;

  if ((decomp = parse_document(xml)) == 0) {
    fprintf(stderr, "error creating decomp\n");
    return 0;
  }
  
  if ((darray = parti_ialign(decomp, igcl, igcu)) == 0) {
    fprintf(stderr, "error creating descriptor\n");
    return 0;
  }
/*   print_darray(darray); */

  cnt = parti_local_size(darray, size);
  A = (int*)malloc(cnt*sizeof(int));
  sts = PARTI_myproc();
  fprintf(ofp, "PROC: %d\n", sts);

  /* ghost_sched_vector */
  elemset('i', A, sts, decomp->rank, size, 0, 0);  
  if ((sched = parti_ghost_sched_vector(darray, fill)) == 0) {
    fprintf(stderr, "error creating schedule\n");
    return 0;
  }
  parti_data_move_int(A, sched, A);
  if (!verify_move(darray, A, fill, 0)) {
    fprintf(ofp, "vector verification failed\n");
    for (i = 0; i < decomp->rank; ++i) {
      fprintf(ofp, "igcl[%d]=%d, igcu[%d]=%d, fill[%d]=%d\n", i, igcl[i], i, igcu[i], i, fill[i]);
    }
    fprint_array(ofp, 'i', A, decomp->rank, size, 0, 0);  
  } else {
    fprintf(ofp, "vector verification succeeded\n");
  }
  parti_free_sched(sched);
  
  /* ghost_sched_span */
  elemset('i', A, sts, decomp->rank, size, 0, 0); 
  if ((sched = parti_ghost_sched_span(darray, fill)) == 0) {
    fprintf(stderr, "error creating schedule\n");
    return 0;
  }
  parti_data_move_int(A, sched, A);
  if (!verify_move(darray, A, fill, 1)) {
    fprintf(ofp, "span verification failed\n");
    for (i = 0; i < decomp->rank; ++i) {
      fprintf(ofp, "igcl[%d]=%d, igcu[%d]=%d, fill[%d]=%d\n", i, igcl[i], i, igcu[i], i, fill[i]);
    }
    fprint_array(ofp, 'i', A, decomp->rank, size, 0, 0);
  } else {
    fprintf(ofp, "span verification succeeded\n");
  }
  parti_free_sched(sched);
  
  /* ghost_sched_all */
  elemset('i', A, sts, decomp->rank, size, 0, 0); 
  if ((sched = parti_ghost_sched_all(darray)) == 0) {
    fprintf(stderr, "error creating schedule\n");
    return 0;
  }
  parti_data_move_int(A, sched, A);
  if (!verify_move(darray, A, fill, 2)) {
    fprintf(ofp, "all verification failed\n");
    for (i = 0; i < decomp->rank; ++i) {
      fprintf(ofp, "igcl[%d]=%d, igcu[%d]=%d, fill[%d]=%d\n", i, igcl[i], i, igcu[i], i, fill[i]);
    }
    fprint_array(ofp, 'i', A, decomp->rank, size, 0, 0);
  } else {
    fprintf(ofp, "all verification succeeded\n");
  }
  parti_free_sched(sched);

  free(A);
  parti_free_darray(darray);
  parti_free_decomp(decomp);

  return 1;
}

int create_params(FILE* ofp, char* xml, int rank, int dim, int* igcl, int* igcu, int* fill) {
  int i, sts;

/*   int lo[3] = {0,0,0}; */
/*   int up[3] = {0,0,1}; */
/*   int gc[3] = {0,0,0}; */
/*   sts = intra_test(ofp, decomp, lo, up, gc); */
/*   PARTI_gsync(); */
/*   return sts; */

  if (dim == rank) {
    if (PARTI_myproc() == 0) {
      for (i = 0; i < rank; ++i) {
	printf("igcl[%d]=%d, igcu[%d]=%d, fill[%d]=%d\n", i, igcl[i], i, igcu[i], i, fill[i]);
      }
      printf("\n");
    }
    sts = intra_test(ofp, xml, igcl, igcu, fill);
    PARTI_gsync();
    return sts;
  }
  
  for (igcl[dim] = 0; igcl[dim] < rank; ++igcl[dim]) {
    for (igcu[dim] = 0; igcu[dim] < rank; ++igcu[dim]) {
      for (fill[dim] = -igcl[dim]; fill[dim] <= igcu[dim]; ++fill[dim]) {
	if (create_params(ofp, xml, rank, dim+1, igcl, igcu, fill) == 0)
	  return 0;
      }
    }
  }
}

void parti_test(FILE* ifp, FILE* ofp, int p, int r) {
  int igcl[MAX_DIM], igcu[MAX_DIM], fill[MAX_DIM];
  char xml[80];
  parti_decomp_t* decomp;
  int sts, test = 0;

  while (fscanf(ifp, "%s", xml) == 1) {
    if (!(decomp = parse_document(xml))) {
      fprintf(stderr, "decomp creation failed\n");
      return;
    }
    
    memset(igcl, 0, sizeof(int)*MAX_DIM);
    memset(igcu, 0, sizeof(int)*MAX_DIM);
    memset(fill, 0, sizeof(int)*MAX_DIM);
    printf("running test %d...", test);
    sts = create_params(ofp, xml, decomp->rank, 0, igcl, igcu, fill);
    if (sts)
      printf("succeeded\n");
    else 
      printf("failed\n");
    parti_free_decomp(decomp);
    ++test;
  }
}
