#include <mpi.h>
#include <stdio.h>
#include <math.h>
#include <bsparti.h>
#include "tests.h"

int main(int argc, char* argv[]) {
  MPI_Comm GROUPCOMM;
  int d, s, c, b, i, p, r;
  int u[3], v[3], w[3][SAMPLES]; /* timings */

  int ds[MAX_DIM] = {0,0,0,0,0,0};
  int gc[MAX_DIM] = {0,0,0,0,0,0};

  int lo[MAX_DIM] = {0,0,0,0,0,0};
  int up[MAX_DIM] = {0,0,0,0,0,0};
  int st[MAX_DIM] = {1,1,1,1,1,1};

  char filename[80];
  FILE* fp;

  if (argc < 4)
    return 1;
  
  /* initialize MPI */
  MPI_Init(&argc, &argv);
  MPI_Comm_dup(MPI_COMM_WORLD, &GROUPCOMM);
  MPI_Comm_rank(GROUPCOMM, &r);
  MPI_Comm_size(GROUPCOMM, &p);
  MPI_Barrier(GROUPCOMM);
  
  /* initialize MBP */
  PARTI_setup(&GROUPCOMM);

  d = atoi(argv[1]); /* dimension */
  s = atoi(argv[2]); /* array size */
  c = atoi(argv[3]); /* ghost cells */
  b = atoi(argv[4]); /* block size */
  
  /* run tests */
  sprintf(filename, "dims%d-%d-%d-%d.dat", VERSION, p, s, c);
  only(r, fp = fopen(filename, "a"));
  fill(i, ds, s);
  fill(i, gc, c);
  only(r, printf("Running test with %d procs and a %d by %dD array with %d ghost cell(s)\n", p, s, d, c));
  for (i = 0; i < SAMPLES; ++i) {
    if (!case1(d, ds, gc, gc, p, r, u)) {
      MPI_Abort(GROUPCOMM, 0);
      return 1;
    }
    MPI_Reduce(u, v, 3, MPI_INT, MPI_SUM, 0, GROUPCOMM);
    only(r, printf("round %d %dus %dus %dus\n", i, v[0], v[1], v[2]));
    w[0][i] = v[0]; w[1][i] = v[1]; w[2][i] = v[2];
  }
  only(r, fprintf(fp, "%d %d %d %d %d %d %d %d %d %d\n", d,
		  avg(w[0]), avg(w[1]), avg(w[2]),
		  min(w[0]), min(w[1]), min(w[2]),
		  max(w[0]), max(w[1]), max(w[2])));
  only(r, fclose(fp));
  
  fill(i, lo, (s - b)/2 - 1);
  fill(i, up, (s + b)/2 - 2);
  sprintf(filename, "dims%d-%d-%d-%d.dat", VERSION, p, s, b);
  only(r, fp = fopen(filename, "a"));
  only(r, printf("Running test with %d procs and a %d by %dD array with a %d by %dD block\n", p, s, d, b, d));
  for (i = 0; i < SAMPLES; ++i) {
    if (!case2(d, ds, lo, up, st, p, r, u)) {
      MPI_Abort(GROUPCOMM, 0);
      return 1;
    }
    MPI_Reduce(u, v, 3, MPI_INT, MPI_SUM, 0, GROUPCOMM);
    only(r, printf("round %d %dus %dus %dus\n", i, v[0], v[1], v[2]));
    w[0][i] = v[0]; w[1][i] = v[1]; w[2][i] = v[2];
  }
  only(r, fprintf(fp, "%d %d %d %d %d %d %d %d %d %d\n", d,
		  avg(w[0]), avg(w[1]), avg(w[2]),
		  min(w[0]), min(w[1]), min(w[2]),
		  max(w[0]), max(w[1]), max(w[2])));
  only(r, fclose(fp));

  /* clean up */
  PARTI_cleanup();
  MPI_Barrier(GROUPCOMM);
  MPI_Comm_free(&GROUPCOMM);
  MPI_Finalize();

  return 0;
}
