#include <bsparti.h>
#include <sys/time.h>

char bd[MAX_DIM] = "BBBBBB"; /* block-cyclic decomposition type */
int ad[MAX_DIM] = {0,1,2,3,4,5};   /* array dimensions */

int egl[MAX_DIM] = {0,0,0,0,0,0};
int egr[MAX_DIM] = {0,0,0,0,0,0};

int ef[MAX_DIM] = {0,0,0,0,0,0}; /* extra flags */
int dd[MAX_DIM] = {0,1,2,3,4,5}; /* decomp dimension */

int case1(int dim, int ds[], int igl[], int igr[], int p, int r, int sts[]) {
  struct timeval bt, et;

  VPROC* vp;
  DECOMP* dc;
  DARRAY* da;

  int local_sizes[MAX_DIM];
  int fill_vector[MAX_DIM];

  int* A;
  int s, t, i;
  
  SCHED* sc;

  /* create descriptor */
  gettimeofday(&bt, 0);
  vp = vProc(1, &p);
  dc = create_decomp(dim, ds);
  embed(dc, vp, 0, p-1);
  distribute(dc, bd);
  da = align(dc, dim, ad, ds, igl, igr, egl, egr, ef, dd);
  gettimeofday(&et, 0);
  sts[0] = (et.tv_sec*1000000 + et.tv_usec) - (bt.tv_sec*1000000 + bt.tv_usec);

  /* create local array */
  laSizes(da, local_sizes);
  s = 1;
  for (i = 0; i < dim; ++i)
    s *= local_sizes[i];
  if (!(A = (int*)calloc(s, sizeof(int)))) {
    perror("calloc");
    return 0;
  }
  
  /* create fill schedules */
  gettimeofday(&bt, 0);
  sc = ghostFillAllSched(da);
  gettimeofday(&et, 0);
  sts[1] = (et.tv_sec*1000000 + et.tv_usec) - (bt.tv_sec*1000000 + bt.tv_usec);
   
  /* move the data */
  gettimeofday(&bt, 0);
  iDataMove(A, sc, A);
  gettimeofday(&et, 0);  
  sts[2] = (et.tv_sec*1000000 + et.tv_usec) - (bt.tv_sec*1000000 + bt.tv_usec);

  /* clean up */
  free_sched(sc);
  delete_DARRAY(da);
  delete_DECOMP(dc);
  delete_VPROC(vp);
  cleanup_after_PARTI();
  free(A);

  return 1;
}

int case2(int dim, int ds[], int lo[], int up[], int st[], int p, int r, int sts[]) {
  struct timeval bt, et;

  int igl[MAX_DIM] = {0,0,0,0,0,0};
  int igr[MAX_DIM] = {0,0,0,0,0,0};
  
  VPROC* vp;
  DECOMP* dc1, *dc2;
  DARRAY* da1, *da2; 
  
  int local_sizes[MAX_DIM];
  int array_dims[MAX_DIM];
  
  int* A, *B;
  int s, t, i;

  SCHED* sc;

  /* create descriptors */
  gettimeofday(&bt, 0);
  vp = vProc(1, &p);
  dc1 = create_decomp(dim, ds);
  dc2 = create_decomp(dim, ds);
  embed(dc1, vp, 0, p-1);
  embed(dc2, vp, 0, p-1);
  distribute(dc1, bd);
  distribute(dc2, bd);
  da1 = align(dc1, dim, ad, ds, igl, igr, egl, egr, ef, dd);
  da2 = align(dc2, dim, ad, ds, igl, igr, egl, egr, ef, dd);
  gettimeofday(&et, 0);
  sts[0] = (et.tv_sec*1000000 + et.tv_usec) - (bt.tv_sec*1000000 + bt.tv_usec);
  
  /* create local arrays */
  laSizes(da1, local_sizes);
  s = 1;
  for (i = 0; i < dim; ++i)
    s *= local_sizes[i];
  if (!(A = (int*)calloc(s, sizeof(int)))) {
    perror("calloc");
    return 0;
  }
  laSizes(da2, local_sizes);
  s = 1;
  for (i = 0; i < dim; ++i)
    s *= local_sizes[i];
  if (!(B = (int*)calloc(s, sizeof(int)))) {
    perror("calloc");
    return 0;
  }

  /* create move schedule */
  gettimeofday(&bt, 0);
  for (i = 0; i < dim; ++i)
    array_dims[i] = i;
  sc = subArraySched(da1, da2, dim, array_dims, lo, up, st, array_dims, lo, up, st); 
  gettimeofday(&et, 0);
  sts[1] = (et.tv_sec*1000000 + et.tv_usec) - (bt.tv_sec*1000000 + bt.tv_usec);
   
  /* move the data */
  gettimeofday(&bt, 0);
  iDataMove(A, sc, B);
  gettimeofday(&et, 0);  
  sts[2] = (et.tv_sec*1000000 + et.tv_usec) - (bt.tv_sec*1000000 + bt.tv_usec);
  
  /* clean up */
  free_sched(sc);
  delete_DARRAY(da1);
  delete_DARRAY(da2);
  delete_DECOMP(dc1);
  delete_DECOMP(dc2);
  delete_VPROC(vp);
  cleanup_after_PARTI();
  free(A);
  free(B);

  return 1;  
}
