#include "ctest.h"
#include <bsparti.h>
#include <stdlib.h>

struct align_params {
  int ad[DIM];
  int ds[DIM];
  int igl[DIM];
  int igr[DIM];
  int egl[DIM];
  int egr[DIM];
  int ef[DIM];
  int dd[DIM];
};

struct trans_params {
  int sd[DIM];
  int sl[DIM];
  int sh[DIM];
  int ss[DIM];
  int dd[DIM];
  int dl[DIM];
  int dh[DIM];
  int ds[DIM];  
};

/* 
 * print_array
 */
void print_array(FILE* fp, int* A, int sizes[]) {
  int i, j;
  
  for (i = 0; i < sizes[0]; ++i) {
    for (j = 0; j < sizes[1]; ++j) {
      fprintf(fp, "%3d ", *(A+i*sizes[1]+j));
    }
    fprintf(fp, "\n");
  }
}

void test1(FILE* fp, int p, int r, struct align_params* ap, struct trans_params* tp) {
  int* A, *B;
  int i, j;

  VPROC* vp;
  DECOMP* dc, *sd;
  DARRAY* da, *sa;
  SCHED* sc;

  int local_sizes[DIM];  
  int fill_vector[DIM];

  /* test array building functions */
  vp = vProc(1, &p);
  dc = create_decomp(DIM, ap->ds);
  embed(dc, vp, 0, p-1);
  distribute(dc, "BB");
  da = align(dc, DIM, ap->ad, ap->ds, ap->igl, ap->igr, ap->egl, ap->egr, ap->ef, ap->dd);
  laSizes(da, local_sizes);
  fprintf(fp, "Local sizes: %d x %d\n", local_sizes[0], local_sizes[1]);

  /* initialize memory */
  A = (int*)calloc(local_sizes[0]*local_sizes[1], sizeof(int));
  for (i = 0; i < local_sizes[0]; ++i) {
    for (j = 0; j < local_sizes[1]; ++j) {
      *(A+i*local_sizes[1]+j) = 100*r + 10*i + j;
    }
  }

  /* print, send internal ghost cells, and print again */
  /*fprintf(stdout, "RANK: %d\n", r);*/
  print_array(fp, A, local_sizes);
  fprintf(fp, "DIMX: %d -> %d\n", lalbnd(da, 0, 0, 1), laubnd(da, 0, DIMX, 1));
  fprintf(fp, "DIMY: %d -> %d\n", lalbnd(da, 1, 0, 1), laubnd(da, 1, DIMY, 1));

  j = 0;
  for (i = 0; i < DIM; ++i) {
    j += ap->igl[i] + ap->igr[i];    
  }

  switch (j % 3) {
  case 0:
    fprintf(fp, "ghostFillAllSched\n");
    sc = ghostFillAllSched(da);
    iDataMove(A, sc, A);
    free_sched(sc);
    break;
  case 1:
    for (i = 0; i < DIM; ++i) {
      for (j = 0; j < DIM; ++j) {
	fill_vector[j] = -ap->igl[j];
      }
      fprintf(fp, "ghostFillSpanSched\n");
      sc = ghostFillSpanSched(da, DIM, fill_vector);
      iDataMove(A, sc, A);
      free_sched(sc);
      for (j = 0; j < DIM; ++j) {
	fill_vector[j] = ap->igr[j];
      }
      fprintf(fp, "ghostFillSpanSched\n");
      sc = ghostFillSpanSched(da, DIM, fill_vector);
      iDataMove(A, sc, A);
      free_sched(sc);
    }
    break;
  case 2:
    for (i = 0; i < DIM; ++i) {
      fprintf(fp, "exchSched\n");
      sc = exchSched(da, i, -ap->igl[i]);
      iDataMove(A, sc, A);
      free_sched(sc);
      fprintf(fp, "exchSched\n");
      sc = exchSched(da, i, ap->igr[i]);
      iDataMove(A, sc, A);
      free_sched(sc);
    }
    break;
  }

  /*fprintf(fp, "RANK: %d\n", r);*/
  print_array(fp, A, local_sizes);

  /* create second array */
  sd = create_decomp(DIM, ap->ds);
  embed(sd, vp, 0, p-1);
  distribute(sd, "BB");
  sa = align(sd, DIM, ap->ad, ap->ds, ap->igl, ap->igr, ap->egl, ap->egr, ap->ef, ap->dd);
  laSizes(sa, local_sizes);
  fprintf(fp, "Local sizes: %d x %d\n", local_sizes[0], local_sizes[1]);
  B = (int*)calloc(local_sizes[0]*local_sizes[1], sizeof(int));
  for (i = 0; i < local_sizes[0]; ++i) {
    for (j = 0; j < local_sizes[1]; ++j) {
      *(B+i*local_sizes[1]+j) = 100*r + 10*i + j;
    }
  }
  
  /* move data between arrays */
  sc = subArraySched(sa, da, DIM, tp->sd, tp->sl, tp->sh, tp->ss, tp->dd, tp->dl, tp->dh, tp->ds);
  iDataMove(A, sc, B);
  free_sched(sc);
  /*fprintf(fp, "RANK: %d\n", r);*/
  print_array(fp, B, local_sizes);

  /* clean up */
  delete_DARRAY(da);
  delete_DARRAY(sa);
  delete_DECOMP(dc);
  delete_DECOMP(sd);
  delete_VPROC(vp);
  cleanup_after_PARTI();
  free(B);
  free(A);
}

/*
 * PARTI_test
 */
void PARTI_test(FILE* fp, int p, int r) {
  int i, j, k, l;

  struct align_params ap;
  struct trans_params tp;

  /* set defaults for array building*/
  ap.ds[0] = DIMX;
  ap.ds[1] = DIMY;
  ap.ad[0] = 0;
  ap.ad[1] = 1;
  ap.igl[0] = 0;
  ap.igl[1] = 0;
  ap.igr[0] = 0;
  ap.igr[1] = 0;
  ap.egl[0] = 0;
  ap.egl[1] = 0;
  ap.egr[0] = 0;
  ap.egr[1] = 0;
  ap.ef[0] = 0;
  ap.ef[1] = 0;
  ap.dd[0] = 0;
  ap.dd[1] = 1;

  /* initialize settings for inter-array transfer */
  tp.sd[0] = 0;
  tp.sd[1] = 1;
  tp.sl[0] = 0;
  tp.sl[1] = 0;
  tp.sh[0] = 2;
  tp.sh[1] = 2;
  tp.ss[0] = 1;
  tp.ss[1] = 1;
  tp.dd[0] = 0;
  tp.dd[1] = 1;
  tp.dl[0] = 5;
  tp.dl[1] = 5;
  tp.dh[0] = 7;
  tp.dh[1] = 7;
  tp.ds[0] = 1;
  tp.ds[1] = 1;
  
  /* vary internal ghost cells */
  for (i = 0; i < 3; ++i) {
    for (j = 0; j < 3; ++j) {
      for (k = 0; k < 3; ++k) {
	for (l = 0; l < 3; ++l) {
	  ap.igl[0] = i;
	  ap.igl[1] = j;
	  ap.igr[0] = k;
	  ap.igr[1] = l;
	  printf("ROUND %d.%d.%d.%d\n", i, j, k, l);
	  test1(fp, p, r, &ap, &tp);
	  PARTI_gsync();
	}
      }
    }
  }
}
